"""
Tests of neo.io.exampleio
"""

import warnings

import unittest

import quantities as pq
import numpy as np

from neo.io.nestio import ColumnIO
from neo.io.nestio import NestIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestNestIO_Analogsignals(BaseTestIO, unittest.TestCase):
    ioclass = NestIO
    entities_to_download = ["nest"]
    entities_to_test = []

    def test_read_analogsignal(self):
        """
        Tests reading files in the 2 different formats:
        - with GIDs, with times as floats
        - with GIDs, with time as integer
        """
        filename = self.get_local_path("nest/0gid-1time-2gex-3Vm-1261-0.dat")
        r = NestIO(filenames=filename)
        r.read_analogsignal(
            gid=1,
            t_stop=1000.0 * pq.ms,
            sampling_period=pq.ms,
            lazy=False,
            id_column=0,
            time_column=1,
            value_column=2,
            value_type="V_m",
        )
        r.read_segment(
            gid_list=[1],
            t_stop=1000.0 * pq.ms,
            sampling_period=pq.ms,
            lazy=False,
            id_column_dat=0,
            time_column_dat=1,
            value_columns_dat=2,
            value_types="V_m",
        )

        filename = self.get_local_path("nest/0gid-1time_in_steps-2Vm-1263-0.dat")
        r = NestIO(filenames=filename)
        r.read_analogsignal(
            gid=1,
            t_stop=1000.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            sampling_period=pq.ms,
            lazy=False,
            id_column=0,
            time_column=1,
            value_column=2,
            value_type="V_m",
        )
        r.read_segment(
            gid_list=[1],
            t_stop=1000.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            sampling_period=pq.ms,
            lazy=False,
            id_column_dat=0,
            time_column_dat=1,
            value_columns_dat=2,
            value_types="V_m",
        )

        filename = self.get_local_path("nest/0gid-1time-2Vm-1259-0.dat")
        r = NestIO(filenames=filename)
        r.read_analogsignal(
            gid=1,
            t_stop=1000.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            sampling_period=pq.ms,
            lazy=False,
            id_column=0,
            time_column=1,
            value_column=2,
            value_type="V_m",
        )
        r.read_segment(
            gid_list=[1],
            t_stop=1000.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            sampling_period=pq.ms,
            lazy=False,
            id_column_dat=0,
            time_column_dat=1,
            value_columns_dat=2,
            value_types="V_m",
        )

    def test_id_column_none_multiple_neurons(self):
        """
        Tests if function correctly raises an error if the user tries to read
        from a file which does not contain unit IDs, but data for multiple
        units.
        """
        filename = self.get_local_path("nest/0time-1255-0.gdf")
        r = NestIO(filenames=filename)
        with self.assertRaises(ValueError):
            r.read_analogsignal(
                t_stop=1000.0 * pq.ms, lazy=False, sampling_period=pq.ms, id_column=None, time_column=0, value_column=1
            )
            r.read_segment(
                t_stop=1000.0 * pq.ms, lazy=False, sampling_period=pq.ms, id_column_gdf=None, time_column_gdf=0
            )

    def test_values(self):
        """
        Tests if the function returns the correct values.
        """
        filename = self.get_local_path("nest/0gid-1time-2gex-3Vm-1261-0.dat")
        id_to_test = 1
        r = NestIO(filenames=filename)
        seg = r.read_segment(
            gid_list=[id_to_test],
            t_stop=1000.0 * pq.ms,
            sampling_period=pq.ms,
            lazy=False,
            id_column_dat=0,
            time_column_dat=1,
            value_columns_dat=2,
            value_types="V_m",
        )

        dat = np.loadtxt(filename)
        target_data = dat[:, 2][np.where(dat[:, 0] == id_to_test)]
        target_data = target_data[:, None]
        st = seg.analogsignals[0]
        np.testing.assert_array_equal(st.magnitude, target_data)

    def test_read_segment(self):
        """
        Tests if signals are correctly stored in a segment.
        """
        filename = self.get_local_path("nest/0gid-1time-2gex-1262-0.dat")
        r = NestIO(filenames=filename)

        id_list_to_test = range(1, 10)
        seg = r.read_segment(
            gid_list=id_list_to_test,
            t_stop=1000.0 * pq.ms,
            sampling_period=pq.ms,
            lazy=False,
            id_column_dat=0,
            time_column_dat=1,
            value_columns_dat=2,
            value_types="V_m",
        )

        self.assertTrue(len(seg.analogsignals) == len(id_list_to_test))

        id_list_to_test = []
        seg = r.read_segment(
            gid_list=id_list_to_test,
            t_stop=1000.0 * pq.ms,
            sampling_period=pq.ms,
            lazy=False,
            id_column_dat=0,
            time_column_dat=1,
            value_columns_dat=2,
            value_types="V_m",
        )

        self.assertEqual(len(seg.analogsignals), 50)

    def test_read_block(self):
        """
        Tests if signals are correctly stored in a block.
        """
        filename = self.get_local_path("nest/0gid-1time-2gex-1262-0.dat")
        r = NestIO(filenames=filename)

        id_list_to_test = range(1, 10)
        blk = r.read_block(
            gid_list=id_list_to_test,
            t_stop=1000.0 * pq.ms,
            sampling_period=pq.ms,
            lazy=False,
            id_column_dat=0,
            time_column_dat=1,
            value_columns_dat=2,
            value_types="V_m",
        )

        self.assertTrue(len(blk.segments[0].analogsignals) == len(id_list_to_test))

    def test_wrong_input(self):
        """
        Tests two cases of wrong user input, namely
        - User does not specify a value column
        - User does not make any specifications
        - User does not define sampling_period as a unit
        - User specifies a non-default value type without
          specifying a value_unit
        - User specifies t_start < 1.*sampling_period
        """
        filename = self.get_local_path("nest/0gid-1time-2gex-1262-0.dat")
        r = NestIO(filenames=filename)
        with self.assertRaises(ValueError):
            r.read_segment(t_stop=1000.0 * pq.ms, lazy=False, id_column_dat=0, time_column_dat=1)
        with self.assertRaises(ValueError):
            r.read_segment()
        with self.assertRaises(ValueError):
            r.read_segment(
                gid_list=[1],
                t_stop=1000.0 * pq.ms,
                sampling_period=1.0 * pq.ms,
                lazy=False,
                id_column_dat=0,
                time_column_dat=1,
                value_columns_dat=2,
                value_types="V_m",
            )

        with self.assertRaises(ValueError):
            r.read_segment(
                gid_list=[1],
                t_stop=1000.0 * pq.ms,
                sampling_period=pq.ms,
                lazy=False,
                id_column_dat=0,
                time_column_dat=1,
                value_columns_dat=2,
                value_types="U_mem",
            )

    def test_t_start_t_stop(self):
        """
        Test for correct t_start and t_stop values of AnalogSignalArrays.
        """
        filename = self.get_local_path("nest/0gid-1time-2gex-1262-0.dat")
        r = NestIO(filenames=filename)

        t_start_targ = 450.0 * pq.ms
        t_stop_targ = 480.0 * pq.ms

        seg = r.read_segment(
            gid_list=[],
            t_start=t_start_targ,
            t_stop=t_stop_targ,
            lazy=False,
            id_column_dat=0,
            time_column_dat=1,
            value_columns_dat=2,
            value_types="V_m",
        )
        anasigs = seg.analogsignals
        for anasig in anasigs:
            self.assertTrue(anasig.t_start == t_start_targ)
            self.assertTrue(anasig.t_stop == t_stop_targ)

    def test_notimeid(self):
        """
        Test for warning, when no time column id was provided.
        """
        filename = self.get_local_path("nest/0gid-1time-2gex-1262-0.dat")
        r = NestIO(filenames=filename)

        t_start_targ = 450.0 * pq.ms
        t_stop_targ = 460.0 * pq.ms
        sampling_period = pq.CompoundUnit("5*ms")

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            seg = r.read_segment(
                gid_list=[],
                t_start=t_start_targ,
                sampling_period=sampling_period,
                t_stop=t_stop_targ,
                lazy=False,
                id_column_dat=0,
                time_column_dat=None,
                value_columns_dat=2,
                value_types="V_m",
            )
            # Verify number and content of warning
            self.assertEqual(len(w), 1)
            self.assertIn("no time column id", str(w[0].message))
        sts = seg.analogsignals
        for st in sts:
            self.assertTrue(st.t_start == 1 * 5 * pq.ms)
            self.assertTrue(st.t_stop == len(st) * sampling_period + 1 * 5 * pq.ms)

    def test_multiple_value_columns(self):
        """
        Test for simultaneous loading of multiple columns from dat file.
        """
        filename = self.get_local_path("nest/0gid-1time-2Vm-3Iex-4Iin-1264-0.dat")
        r = NestIO(filenames=filename)

        sampling_period = pq.CompoundUnit("5*ms")
        seg = r.read_segment(gid_list=[1001], value_columns_dat=[2, 3], sampling_period=sampling_period)
        anasigs = seg.analogsignals
        self.assertEqual(len(anasigs), 2)

    def test_single_gid(self):
        filename = self.get_local_path("nest/N1-0gid-1time-2Vm-1265-0.dat")
        r = NestIO(filenames=filename)
        anasig = r.read_analogsignal(
            gid=1,
            t_stop=1000.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            sampling_period=pq.ms,
            lazy=False,
            id_column=0,
            time_column=1,
            value_column=2,
            value_type="V_m",
        )
        assert anasig.annotations["id"] == 1

    def test_no_gid(self):
        filename = self.get_local_path("nest/N1-0time-1Vm-1266-0.dat")
        r = NestIO(filenames=filename)
        anasig = r.read_analogsignal(
            gid=None,
            t_stop=1000.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            sampling_period=pq.ms,
            lazy=False,
            id_column=None,
            time_column=0,
            value_column=1,
            value_type="V_m",
        )
        self.assertEqual(anasig.annotations["id"], None)
        self.assertEqual(len(anasig), 19)

    def test_no_gid_no_time(self):
        filename = self.get_local_path("nest/N1-0Vm-1267-0.dat")
        r = NestIO(filenames=filename)
        anasig = r.read_analogsignal(
            gid=None,
            sampling_period=pq.ms,
            lazy=False,
            id_column=None,
            time_column=None,
            value_column=0,
            value_type="V_m",
        )
        self.assertEqual(anasig.annotations["id"], None)
        self.assertEqual(len(anasig), 19)


class TestNestIO_Spiketrains(BaseTestIO, unittest.TestCase):
    ioclass = NestIO
    entities_to_download = ["nest"]

    def test_read_spiketrain(self):
        """
        Tests reading files in the 4 different formats:
        - without GIDs, with times as floats
        - without GIDs, with times as integers in time steps
        - with GIDs, with times as floats
        - with GIDs, with times as integers in time steps
        """
        filename = self.get_local_path("nest/0time-1255-0.gdf")
        r = NestIO(filenames=filename)
        r.read_spiketrain(t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms, lazy=False, id_column=None, time_column=0)
        r.read_segment(t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms, lazy=False, id_column_gdf=None, time_column_gdf=0)

        filename = self.get_local_path("nest/0time_in_steps-1257-0.gdf")
        r = NestIO(filenames=filename)
        r.read_spiketrain(
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            lazy=False,
            id_column=None,
            time_column=0,
        )
        r.read_segment(
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            lazy=False,
            id_column_gdf=None,
            time_column_gdf=0,
        )

        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        r.read_spiketrain(
            gdf_id=1, t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms, lazy=False, id_column_gdf=0, time_column_gdf=1
        )
        r.read_segment(
            gid_list=[1], t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms, lazy=False, id_column_gdf=0, time_column_gdf=1
        )

        filename = self.get_local_path("nest/0gid-1time_in_steps-1258-0.gdf")
        r = NestIO(filenames=filename)
        r.read_spiketrain(
            gdf_id=1,
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            lazy=False,
            id_column=0,
            time_column=1,
        )
        r.read_segment(
            gid_list=[1],
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            lazy=False,
            id_column_gdf=0,
            time_column_gdf=1,
        )

    def test_read_integer(self):
        """
        Tests if spike times are actually stored as integers if they are stored
        in time steps in the file.
        """
        filename = self.get_local_path("nest/0time_in_steps-1257-0.gdf")
        r = NestIO(filenames=filename)
        st = r.read_spiketrain(
            gdf_id=None,
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            lazy=False,
            id_column=None,
            time_column=0,
        )
        self.assertTrue(st.magnitude.dtype == np.int32)
        seg = r.read_segment(
            gid_list=[None],
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            lazy=False,
            id_column_gdf=None,
            time_column_gdf=0,
        )
        sts = seg.spiketrains
        self.assertTrue(all([st.magnitude.dtype == np.int32 for st in sts]))

        filename = self.get_local_path("nest/0gid-1time_in_steps-1258-0.gdf")
        r = NestIO(filenames=filename)
        st = r.read_spiketrain(
            gdf_id=1,
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            lazy=False,
            id_column=0,
            time_column=1,
        )
        self.assertTrue(st.magnitude.dtype == np.int32)
        seg = r.read_segment(
            gid_list=[1],
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            time_unit=pq.CompoundUnit("0.1*ms"),
            lazy=False,
            id_column_gdf=0,
            time_column_gdf=1,
        )
        sts = seg.spiketrains
        self.assertTrue(all([st.magnitude.dtype == np.int32 for st in sts]))

    def test_read_float(self):
        """
        Tests if spike times are stored as floats if they
        are stored as floats in the file.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        st = r.read_spiketrain(
            gdf_id=1, t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms, lazy=False, id_column=0, time_column=1
        )
        self.assertTrue(st.magnitude.dtype == float)
        seg = r.read_segment(
            gid_list=[1], t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms, lazy=False, id_column_gdf=0, time_column_gdf=1
        )
        sts = seg.spiketrains
        self.assertTrue(all([s.magnitude.dtype == np.float64 for s in sts]))

    def test_values(self):
        """
        Tests if the routine loads the correct numbers from the file.
        """
        id_to_test = 1
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        seg = r.read_segment(
            gid_list=[id_to_test],
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            lazy=False,
            id_column_gdf=0,
            time_column_gdf=1,
        )

        dat = np.loadtxt(filename)
        target_data = dat[:, 1][np.where(dat[:, 0] == id_to_test)]

        st = seg.spiketrains[0]
        np.testing.assert_array_equal(st.magnitude, target_data)

    def test_read_segment(self):
        """
        Tests if spiketrains are correctly stored in a segment.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)

        id_list_to_test = range(1, 10)
        seg = r.read_segment(
            gid_list=id_list_to_test,
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            lazy=False,
            id_column_gdf=0,
            time_column_gdf=1,
        )
        self.assertTrue(len(seg.spiketrains) == len(id_list_to_test))

        id_list_to_test = []
        seg = r.read_segment(
            gid_list=id_list_to_test,
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            lazy=False,
            id_column_gdf=0,
            time_column_gdf=1,
        )
        self.assertTrue(len(seg.spiketrains) == 50)

    def test_read_segment_accepts_range(self):
        """
        Tests if spiketrains can be retrieved by specifying a range of GDF IDs.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)

        seg = r.read_segment(
            gid_list=(10, 39),
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            lazy=False,
            id_column_gdf=0,
            time_column_gdf=1,
        )
        self.assertEqual(len(seg.spiketrains), 30)

    def test_read_segment_range_is_reasonable(self):
        """
        Tests if error is thrown correctly, when second entry is smaller than
        the first one of the range.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)

        seg = r.read_segment(
            gid_list=(10, 10),
            t_start=400.0 * pq.ms,
            t_stop=500.0 * pq.ms,
            lazy=False,
            id_column_gdf=0,
            time_column_gdf=1,
        )
        self.assertEqual(len(seg.spiketrains), 1)
        with self.assertRaises(ValueError):
            r.read_segment(
                gid_list=(10, 9),
                t_start=400.0 * pq.ms,
                t_stop=500.0 * pq.ms,
                lazy=False,
                id_column_gdf=0,
                time_column_gdf=1,
            )

    def test_read_spiketrain_annotates(self):
        """
        Tests if correct annotation is added when reading a spike train.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        ID = 7
        st = r.read_spiketrain(gdf_id=ID, t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms)
        self.assertEqual(ID, st.annotations["id"])

    def test_read_segment_annotates(self):
        """
        Tests if correct annotation is added when reading a segment.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        IDs = (5, 11)
        sts = r.read_segment(gid_list=(5, 11), t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms)
        for ID in np.arange(5, 12):
            self.assertEqual(ID, sts.spiketrains[ID - 5].annotations["id"])

    def test_adding_custom_annotation(self):
        """
        Tests if custom annotation is correctly added.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        st = r.read_spiketrain(gdf_id=0, t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms, layer="L23", population="I")
        self.assertEqual(0, st.annotations.pop("id"))
        self.assertEqual("L23", st.annotations.pop("layer"))
        self.assertEqual("I", st.annotations.pop("population"))
        self.assertEqual({}, st.annotations)

    def test_wrong_input(self):
        """
        Tests two cases of wrong user input, namely
        - User does not specify neuron IDs although the file contains IDs.
        - User does not make any specifications.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        with self.assertRaises(ValueError):
            r.read_segment(t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms, lazy=False, id_column_gdf=0, time_column_gdf=1)
        with self.assertRaises(ValueError):
            r.read_segment()

    def test_t_start_t_stop(self):
        """
        Tests if the t_start and t_stop arguments are correctly processed.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)

        t_stop_targ = 490.0 * pq.ms
        t_start_targ = 410.0 * pq.ms

        seg = r.read_segment(
            gid_list=[], t_start=t_start_targ, t_stop=t_stop_targ, lazy=False, id_column_gdf=0, time_column_gdf=1
        )
        sts = seg.spiketrains
        self.assertTrue(
            np.max([np.max(st.magnitude) for st in sts if len(st) > 0])
            < t_stop_targ.rescale(sts[0].times.units).magnitude
        )
        self.assertTrue(
            np.min([np.min(st.magnitude) for st in sts if len(st) > 0])
            >= t_start_targ.rescale(sts[0].times.units).magnitude
        )

    def test_t_start_undefined_raises_error(self):
        """
        Tests if undefined t_start, i.e., t_start=None raises error.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        with self.assertRaises(ValueError):
            r.read_spiketrain(gdf_id=1, t_stop=500.0 * pq.ms, lazy=False, id_column=0, time_column=1)
        with self.assertRaises(ValueError):
            r.read_segment(gid_list=[1, 2, 3], t_stop=500.0 * pq.ms, lazy=False, id_column_gdf=0, time_column_gdf=1)

    def test_t_stop_undefined_raises_error(self):
        """
        Tests if undefined t_stop, i.e., t_stop=None raises error.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        with self.assertRaises(ValueError):
            r.read_spiketrain(gdf_id=1, t_start=400.0 * pq.ms, lazy=False, id_column=0, time_column=1)
        with self.assertRaises(ValueError):
            r.read_segment(gid_list=[1, 2, 3], t_start=400.0 * pq.ms, lazy=False, id_column_gdf=0, time_column_gdf=1)

    def test_gdf_id_illdefined_raises_error(self):
        """
        Tests if ill-defined gdf_id in read_spiketrain (i.e., None, list, or
        empty list) raises error.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        with self.assertRaises(ValueError):
            r.read_spiketrain(gdf_id=[], t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms)
        with self.assertRaises(ValueError):
            r.read_spiketrain(gdf_id=[1], t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms)
        with self.assertRaises(ValueError):
            r.read_spiketrain(t_start=400.0 * pq.ms, t_stop=500.0 * pq.ms)

    def test_read_segment_can_return_empty_spiketrains(self):
        """
        Tests if read_segment makes sure that only non-zero spike trains are
        returned.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        seg = r.read_segment(gid_list=[], t_start=400.4 * pq.ms, t_stop=400.5 * pq.ms)
        for st in seg.spiketrains:
            self.assertEqual(st.size, 0)

    def test_read_spiketrain_can_return_empty_spiketrain(self):
        """
        Tests if read_spiketrain returns an empty SpikeTrain if no spikes are in
        time range.
        """
        filename = self.get_local_path("nest/0gid-1time-1256-0.gdf")
        r = NestIO(filenames=filename)
        st = r.read_spiketrain(gdf_id=0, t_start=400.0 * pq.ms, t_stop=410.0 * pq.ms)
        self.assertEqual(st.size, 0)


class TestNestIO_multiple_signal_types(BaseTestIO, unittest.TestCase):
    ioclass = NestIO
    entities_to_test = []
    entities_to_download = ["nest"]

    def test_read_analogsignal_and_spiketrain(self):
        """
        Test if spiketrains and analogsignals can be read simultaneously
        using read_segment
        """
        files = ["nest/0gid-1time-2gex-3Vm-1261-0.dat", "nest/0gid-1time_in_steps-1258-0.gdf"]
        filenames = [self.get_local_path(e) for e in files]

        r = NestIO(filenames=filenames)
        seg = r.read_segment(
            gid_list=[],
            t_start=400 * pq.ms,
            t_stop=600 * pq.ms,
            id_column_gdf=0,
            time_column_gdf=1,
            id_column_dat=0,
            time_column_dat=1,
            value_columns_dat=2,
        )
        self.assertEqual(len(seg.spiketrains), 50)
        self.assertEqual(len(seg.analogsignals), 50)


class TestColumnIO(BaseTestIO, unittest.TestCase):
    ioclass = NestIO
    entities_to_test = []
    entities_to_download = ["nest"]

    def setUp(self):
        BaseTestIO.setUp(self)
        filename = self.get_local_path("nest/0gid-1time-2Vm-3gex-4gin-1260-0.dat")
        self.testIO = ColumnIO(filename=filename)

    def test_no_arguments(self):
        """
        Test if data can be read using the default keyword arguments.
        """
        columns = self.testIO.get_columns()
        expected = self.testIO.data
        np.testing.assert_array_equal(columns, expected)

    def test_single_column_id(self):
        """
        Test if the column_ids keywords works properly.
        """
        column = self.testIO.get_columns(column_ids=1)
        expected = self.testIO.data[:, [1]]
        np.testing.assert_array_equal(column, expected)

    def test_multiple_column_ids(self):
        """
        Test if multiple columns can be read at the same time.
        """
        columns = self.testIO.get_columns(column_ids=range(2))
        expected = self.testIO.data[:, [0, 1]]
        np.testing.assert_array_equal(columns, expected)

    def test_no_condition(self):
        """
        Test if a missing condition function leads to a warning
        """
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            self.testIO.get_columns(condition_column=0)
            # Verify number and content of warning
            assert len(w) == 1
            assert "no condition" in str(w[-1].message)

    def test_no_condition_column(self):
        """
        Test if a missing condition column leads to an error
        """
        with self.assertRaises(ValueError) as context:
            self.testIO.get_columns(condition=lambda x: True)

        self.assertTrue("no condition_column ID provided" in str(context.exception))

    def test_correct_condition_selection(self):
        """
        Test if combination of condition function and condition_column works
        properly.
        """
        condition_column = 0

        def condition_function(x):
            return x > 10

        result = self.testIO.get_columns(condition=condition_function, condition_column=0)
        selected_ids = np.where(condition_function(self.testIO.data[:, condition_column]))[0]
        expected = self.testIO.data[selected_ids, :]

        np.testing.assert_array_equal(result, expected)

        assert all(condition_function(result[:, condition_column]))

    def test_sorting(self):
        """
        Test if presorting of columns work properly.
        """
        result = self.testIO.get_columns(sorting_columns=0)

        assert len(result) > 0
        assert all(np.diff(result[:, 0]) >= 0)


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.winwcprawio import WinWcpRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestWinWcpRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = WinWcpRawIO
    entities_to_download = ["winwcp"]
    entities_to_test = ["winwcp/File_winwcp_1.wcp"]


if __name__ == "__main__":
    unittest.main()



"""
IgorProIO Demo
===========================

"""

###########################################################
# Import our packages
import os
from urllib.request import urlretrieve
import zipfile
import matplotlib.pyplot as plt
from neo.io import get_io

#############################################################
# Then download some data
# Downloaded from Human Brain Project Collaboratory
# Digital Reconstruction of Neocortical Microcircuitry (nmc-portal)
# http://microcircuits.epfl.ch/#/animal/8ecde7d1-b2d2-11e4-b949-6003088da632


datafile_url = "https://microcircuits.epfl.ch/data/released_data/B95.zip"
filename_zip = "B95.zip"
filename = "grouped_ephys/B95/B95_Ch0_IDRest_107.ibw"
urlretrieve(datafile_url, filename_zip)

zip_ref = zipfile.ZipFile(filename_zip)  # create zipfile object
zip_ref.extract(path=".", member=filename)  # extract file to dir
zip_ref.close()

######################################################
# Once we have our data we can use `get_io` to find an
# io (Igor in this case). Then we read the analogsignals
# Finally we will make some nice plots
reader = get_io(filename)
signal = reader.read_analogsignal()
plt.plot(signal.times, signal)
plt.xlabel(signal.sampling_period.dimensionality)
plt.ylabel(signal.dimensionality)

plt.show()



#
"""
Tests of neo.io.nwbio
"""

import os
import unittest
from datetime import datetime

try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve
from neo.test.iotest.common_io_test import BaseTestIO
from neo.core import AnalogSignal, SpikeTrain, Event, Epoch, IrregularlySampledSignal, Segment, Block

from neo.rawio.examplerawio import ExampleRawIO
from neo.io.proxyobjects import AnalogSignalProxy, SpikeTrainProxy, EventProxy, EpochProxy

try:
    import pynwb
    from neo.io.nwbio import NWBIO

    HAVE_PYNWB = True
except (ImportError, SyntaxError):
    NWBIO = None
    HAVE_PYNWB = False
import quantities as pq
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose


@unittest.skipUnless(HAVE_PYNWB, "requires pynwb")
class TestNWBIO(BaseTestIO, unittest.TestCase):
    ioclass = NWBIO
    entities_to_download = ["nwb"]
    entities_to_test = [
        # Files from Allen Institute:
        "nwb/H19.29.141.11.21.01.nwb",  # 7 MB
    ]

    def test_roundtrip(self):

        annotations = {"session_start_time": datetime.now()}
        # Define Neo blocks
        bl0 = Block(name="First block", **annotations)
        bl1 = Block(name="Second block", **annotations)
        bl2 = Block(name="Third block", **annotations)
        original_blocks = [bl0, bl1, bl2]

        num_seg = 4  # number of segments
        num_chan = 3  # number of channels

        for blk in original_blocks:

            for ind in range(num_seg):  # number of Segments
                seg = Segment(index=ind)
                blk.segments.append(seg)

            for seg in blk.segments:  # AnalogSignal objects

                # 3 Neo AnalogSignals
                a = AnalogSignal(
                    name="Signal_a %s" % (seg.name),
                    signal=np.random.randn(44, num_chan) * pq.nA,
                    sampling_rate=10 * pq.kHz,
                    t_start=50 * pq.ms,
                )
                b = AnalogSignal(
                    name="Signal_b %s" % (seg.name),
                    signal=np.random.randn(64, num_chan) * pq.mV,
                    sampling_rate=8 * pq.kHz,
                    t_start=40 * pq.ms,
                )
                c = AnalogSignal(
                    name="Signal_c %s" % (seg.name),
                    signal=np.random.randn(33, num_chan) * pq.uA,
                    sampling_rate=10 * pq.kHz,
                    t_start=120 * pq.ms,
                )
                # 2 Neo IrregularlySampledSignals
                d = IrregularlySampledSignal(np.arange(7.0) * pq.ms, np.random.randn(7, num_chan) * pq.mV)

                # 2 Neo SpikeTrains
                train = SpikeTrain(times=[1, 2, 3] * pq.s, t_start=1.0, t_stop=10.0)
                train2 = SpikeTrain(times=[4, 5, 6] * pq.s, t_stop=10.0)
                # todo: add waveforms

                # 1 Neo Event
                evt = Event(name="Event", times=np.arange(0, 30, 10) * pq.ms, labels=np.array(["ev0", "ev1", "ev2"]))

                # 2 Neo Epochs
                epc = Epoch(
                    times=np.arange(0, 30, 10) * pq.s,
                    durations=[10, 5, 7] * pq.ms,
                    labels=np.array(["btn0", "btn1", "btn2"]),
                )

                epc2 = Epoch(
                    times=np.arange(10, 40, 10) * pq.s,
                    durations=[9, 3, 8] * pq.ms,
                    labels=np.array(["btn3", "btn4", "btn5"]),
                )

                seg.spiketrains.append(train)
                seg.spiketrains.append(train2)

                seg.epochs.append(epc)
                seg.epochs.append(epc2)

                seg.analogsignals.append(a)
                seg.analogsignals.append(b)
                seg.analogsignals.append(c)
                seg.irregularlysampledsignals.append(d)
                seg.events.append(evt)

        # write to file
        test_file_name = "test_round_trip.nwb"
        iow = NWBIO(filename=test_file_name, mode="w")
        iow.write_all_blocks(original_blocks)

        ior = NWBIO(filename=test_file_name, mode="r")
        retrieved_blocks = ior.read_all_blocks()

        self.assertEqual(len(retrieved_blocks), 3)
        self.assertEqual(len(retrieved_blocks[2].segments), num_seg)

        original_signal_22b = original_blocks[2].segments[2].analogsignals[1]
        retrieved_signal_22b = retrieved_blocks[2].segments[2].analogsignals[1]
        for attr_name in ("name", "units", "sampling_rate", "t_start"):
            retrieved_attribute = getattr(retrieved_signal_22b, attr_name)
            original_attribute = getattr(original_signal_22b, attr_name)
            self.assertEqual(retrieved_attribute, original_attribute)
        assert_array_equal(retrieved_signal_22b.magnitude, original_signal_22b.magnitude)

        original_issignal_22d = original_blocks[2].segments[2].irregularlysampledsignals[0]
        retrieved_issignal_22d = retrieved_blocks[2].segments[2].irregularlysampledsignals[0]
        for attr_name in ("name", "units", "t_start"):
            retrieved_attribute = getattr(retrieved_issignal_22d, attr_name)
            original_attribute = getattr(original_issignal_22d, attr_name)
            self.assertEqual(retrieved_attribute, original_attribute)
        assert_array_equal(
            retrieved_issignal_22d.times.rescale("ms").magnitude, original_issignal_22d.times.rescale("ms").magnitude
        )
        assert_array_equal(retrieved_issignal_22d.magnitude, original_issignal_22d.magnitude)

        original_event_11 = original_blocks[1].segments[1].events[0]
        retrieved_event_11 = retrieved_blocks[1].segments[1].events[0]
        for attr_name in ("name",):
            retrieved_attribute = getattr(retrieved_event_11, attr_name)
            original_attribute = getattr(original_event_11, attr_name)
            self.assertEqual(retrieved_attribute, original_attribute)
        assert_array_equal(retrieved_event_11.rescale("ms").magnitude, original_event_11.rescale("ms").magnitude)
        assert_array_equal(retrieved_event_11.labels, original_event_11.labels)

        original_spiketrain_131 = original_blocks[1].segments[1].spiketrains[1]
        retrieved_spiketrain_131 = retrieved_blocks[1].segments[1].spiketrains[1]
        for attr_name in ("name", "t_start", "t_stop"):
            retrieved_attribute = getattr(retrieved_spiketrain_131, attr_name)
            original_attribute = getattr(original_spiketrain_131, attr_name)
            self.assertEqual(retrieved_attribute, original_attribute)
        assert_array_equal(
            retrieved_spiketrain_131.times.rescale("ms").magnitude,
            original_spiketrain_131.times.rescale("ms").magnitude,
        )

        original_epoch_11 = original_blocks[1].segments[1].epochs[0]
        retrieved_epoch_11 = retrieved_blocks[1].segments[1].epochs[0]
        for attr_name in ("name",):
            retrieved_attribute = getattr(retrieved_epoch_11, attr_name)
            original_attribute = getattr(original_epoch_11, attr_name)
            self.assertEqual(retrieved_attribute, original_attribute)
        assert_array_equal(retrieved_epoch_11.rescale("ms").magnitude, original_epoch_11.rescale("ms").magnitude)
        assert_allclose(
            retrieved_epoch_11.durations.rescale("ms").magnitude, original_epoch_11.durations.rescale("ms").magnitude
        )
        assert_array_equal(retrieved_epoch_11.labels, original_epoch_11.labels)
        os.remove(test_file_name)

    def test_roundtrip_with_annotations(self):
        # test with NWB-specific annotations

        original_block = Block(name="experiment", session_start_time=datetime.now())
        segment = Segment(name="session 1")
        original_block.segments.append(segment)

        electrode_annotations = {
            "name": "electrode #1",
            "description": "intracellular electrode",
            "device": {"name": "electrode #1"},
        }
        stimulus_annotations = {
            "nwb_group": "stimulus",
            "nwb_neurodata_type": ("pynwb.icephys", "CurrentClampStimulusSeries"),
            "nwb_electrode": electrode_annotations,
            "nwb:sweep_number": 1,
            "nwb:gain": 1.0,
        }
        response_annotations = {
            "nwb_group": "acquisition",
            "nwb_neurodata_type": ("pynwb.icephys", "CurrentClampSeries"),
            "nwb_electrode": electrode_annotations,
            "nwb:sweep_number": 1,
            "nwb:gain": 1.0,
            "nwb:bias_current": 1e-12,
            "nwb:bridge_balance": 70e6,
            "nwb:capacitance_compensation": 1e-12,
        }
        stimulus = AnalogSignal(
            np.random.randn(100, 1) * pq.nA,
            sampling_rate=5 * pq.kHz,
            t_start=50 * pq.ms,
            name="stimulus",
            **stimulus_annotations,
        )
        response = AnalogSignal(
            np.random.randn(100, 1) * pq.mV,
            sampling_rate=5 * pq.kHz,
            t_start=50 * pq.ms,
            name="response",
            **response_annotations,
        )
        segment.analogsignals = [stimulus, response]
        assert stimulus.segment is response.segment is segment

        test_file_name = "test_round_trip_with_annotations.nwb"
        iow = NWBIO(filename=test_file_name, mode="w")
        iow.write_all_blocks([original_block])

        nwbfile = pynwb.NWBHDF5IO(test_file_name, mode="r").read()

        self.assertIsInstance(nwbfile.acquisition[response.name], pynwb.icephys.CurrentClampSeries)
        self.assertIsInstance(nwbfile.stimulus[stimulus.name], pynwb.icephys.CurrentClampStimulusSeries)
        self.assertEqual(nwbfile.acquisition[response.name].bridge_balance, response_annotations["nwb:bridge_balance"])

        ior = NWBIO(filename=test_file_name, mode="r")
        retrieved_block = ior.read_all_blocks()[0]

        original_response = original_block.segments[0].filter(name=response.name)[0]
        retrieved_response = retrieved_block.segments[0].filter(name=response.name)[0]
        for attr_name in ("name", "units", "sampling_rate", "t_start"):
            retrieved_attribute = getattr(retrieved_response, attr_name)
            original_attribute = getattr(original_response, attr_name)
            self.assertEqual(retrieved_attribute, original_attribute)
        assert_array_equal(retrieved_response.magnitude, original_response.magnitude)

        os.remove(test_file_name)

    def test_write_proxy_objects(self):
        test_file_name = self.local_test_dir / "test_round_trip_with_annotations.nwb"

        # generate dummy IO as basis for ProxyObjects
        self.proxy_reader = ExampleRawIO(filename="my_filename.fake")
        self.proxy_reader.parse_header()

        # generate test structure with proxy objects
        original_block = Block(
            name="myblock",
            session_start_time=datetime.now().astimezone(),
            session_description=str(test_file_name),
            identifier=str(test_file_name),
        )
        seg = Segment(name="mysegment")
        original_block.segments.append(seg)

        # create proxy objects
        proxy_anasig = AnalogSignalProxy(
            rawio=self.proxy_reader,
            stream_index=0,
            inner_stream_channels=None,
            block_index=0,
            seg_index=0,
        )
        seg.analogsignals.append(proxy_anasig)

        proxy_sptr = SpikeTrainProxy(rawio=self.proxy_reader, spike_channel_index=0, block_index=0, seg_index=0)
        seg.spiketrains.append(proxy_sptr)

        proxy_event = EventProxy(rawio=self.proxy_reader, event_channel_index=0, block_index=0, seg_index=0)
        seg.events.append(proxy_event)

        proxy_epoch = EpochProxy(rawio=self.proxy_reader, event_channel_index=1, block_index=0, seg_index=0)
        seg.epochs.append(proxy_epoch)

        original_block.check_relationships()

        iow = NWBIO(filename=test_file_name, mode="w")

        # writing data via proxyobjects
        iow.write_all_blocks([original_block])

        # checking written data
        ior = NWBIO(filename=test_file_name, mode="r")
        retrieved_block = ior.read_all_blocks()[0]

        for original_segment, retrieved_segment in zip(original_block.segments, retrieved_block.segments):
            assert_array_equal(
                original_segment.analogsignals[0].load().magnitude, retrieved_segment.analogsignals[0].magnitude
            )
            assert_array_equal(
                original_segment.spiketrains[0].load().magnitude, retrieved_segment.spiketrains[0].magnitude
            )
            assert_array_equal(original_segment.events[0].load().magnitude, retrieved_segment.events[0].magnitude)
            assert_array_equal(original_segment.epochs[0].load().magnitude, retrieved_segment.epochs[0].magnitude)


if __name__ == "__main__":
    if HAVE_PYNWB:
        print("pynwb.__version__ = ", pynwb.__version__)
    unittest.main()



"""
ExampleRawIO is a class of a fake example.
This is to be used when coding a new RawIO.


Rules for creating a new class:
  1. Step 1: Create the main class
    * Create a file in **neo/rawio/** that ends with "rawio.py"
    * Create the class that inherits from BaseRawIO
    * copy/paste all methods that need to be implemented.
    * code hard! The main difficulty is `_parse_header()`.
      In short you have to create a mandatory dict that
      contains channel information::

            self.header = {}
            self.header['nb_block'] = 2
            self.header['nb_segment'] = [2, 3]
            self.header['signal_streams'] = signal_streams
            self.header['signal_channels'] = signal_channels
            self.header['spike_channels'] = spike_channels
            self.header['event_channels'] = event_channels

  2. Step 2: RawIO test:
    * create a file in neo/rawio/tests with the same name with "test_" prefix
    * copy paste neo/rawio/tests/test_examplerawio.py and do the same

  3. Step 3: Create the neo.io class with the wrapper
    * Create a file in neo/io/ that ends with "io.py"
    * Create a class that inherits both your RawIO class and the BaseFromRaw class
    * copy/paste from neo/io/exampleio.py

  4.Step 4: IO test
    * create a file in neo/test/iotest with the same name as previously with "test_" prefix
    * copy/paste from neo/test/iotest/test_exampleio.py

"""

from __future__ import annotations

import numpy as np
from pathlib import Path

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class ExampleRawIO(BaseRawIO):
    """
    Class for "reading" fake data from an imaginary file.

    For the user, it gives access to raw data (signals, event, spikes) as they
    are in the (fake) file int16 and int64.

    For a developer, it is just an example showing guidelines for someone who wants
    to develop a new IO module.

    Two rules for developers:
      * Respect the :ref:`section-neo-rawio-API`
      * Follow the guidelines in :ref:`section-raw-io-recipe`

    This fake IO:
        * has 2 blocks
        * blocks have 2 and 3 segments
        * has  2 signals streams  of 8 channels each (sample_rate = 10000) so 16 channels in total
        * has 3 spike_channels
        * has 2 event channels: one has *type=event*, the other has
          *type=epoch*


    Usage:
        >>> import neo.rawio
        >>> r = neo.rawio.ExampleRawIO(filename='itisafake.nof')
        >>> r.parse_header()
        >>> print(r)
        >>> raw_chunk = r.get_analogsignal_chunk(block_index=0, seg_index=0,
                            i_start=0, i_stop=1024,  channel_names=channel_names)
        >>> float_chunk = reader.rescale_signal_raw_to_float(raw_chunk, dtype='float64',
                            channel_indexes=[0, 3, 6])
        >>> spike_timestamp = reader.spike_timestamps(spike_channel_index=0,
                            t_start=None, t_stop=None)
        >>> spike_times = reader.rescale_spike_timestamp(spike_timestamp, 'float64')
        >>> ev_timestamps, _, ev_labels = reader.event_timestamps(event_channel_index=0)

    """

    extensions = ["fake"]
    rawmode = "one-file"

    def __init__(self, filename: str | Path = ""):
        BaseRawIO.__init__(self)
        # note that this filename is ued in self._source_name
        self.filename = filename

    def _source_name(self):
        # this function is used by __repr__
        # for general cases self.filename is good
        # But for URL you could mask some part of the URL to keep
        # the main part.
        return self.filename

    def _parse_header(self):
        # This is the central part of a RawIO
        # we need to collect from the original format all
        # information required for fast access
        # at any place in the file
        # In short `_parse_header()` can be slow but
        # `_get_analogsignal_chunk()` needs to be as fast as possible

        # create fake signal streams information
        signal_streams = []
        for c in range(2):
            name = f"stream {c}"
            stream_id = c
            signal_streams.append((name, stream_id))
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # create fake signal channels information
        # This is mandatory!!!!
        # gain/offset/units are really important because
        # the scaling to real value will be done with that
        # The real signal will be evaluated as `(raw_signal * gain + offset) * pq.Quantity(units)`
        signal_channels = []
        for c in range(16):
            ch_name = f"ch{c}"
            # our channel id is c+1 just for fun
            # Note that chan_id should be related to
            # original channel id in the file format
            # so that the end user should not be confused when reading datasets
            chan_id = c + 1
            sr = 10000.0  # Hz
            dtype = "int16"
            units = "uV"
            gain = 1000.0 / 2**16
            offset = 0.0
            # stream_id indicates how to group channels
            # channels inside a "stream" share the same characteristics
            #  (sampling rate/dtype/t_start/units/...)
            stream_id = str(c // 8)
            signal_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # A stream can contain signals with different physical units.
        # Here, the two last channels will have different units (pA)
        # Since AnalogSignals must have consistent units across channels,
        # this stream will be split in 2 parts on the neo.io level and finally 3 AnalogSignals
        # will be generated per Segment.
        signal_channels[-2:]["units"] = "pA"

        # create fake unit channels
        # This is mandatory!!!!
        # Note that if there is no waveform at all in the file
        # then wf_units/wf_gain/wf_offset/wf_left_sweep/wf_sampling_rate
        # can be set to any value because _spike_raw_waveforms
        # will return None
        spike_channels = []
        for c in range(3):
            unit_name = f"unit{c}"
            unit_id = f"#{c}"
            wf_units = "uV"
            wf_gain = 1000.0 / 2**16
            wf_offset = 0.0
            wf_left_sweep = 20
            wf_sampling_rate = 10000.0
            spike_channels.append((unit_name, unit_id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate))
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # creating event/epoch channel
        # This is mandatory!!!!
        # In RawIO epoch and event are dealt with in the same way.
        event_channels = []
        event_channels.append(("Some events", "ev_0", "event"))
        event_channels.append(("Some epochs", "ep_1", "epoch"))
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # fill information into the header dict
        # This is mandatory!!!!!
        self.header = {}
        self.header["nb_block"] = 2
        self.header["nb_segment"] = [2, 3]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # insert some annotations/array_annotations at some place
        # at neo.io level. IOs can add annotations
        # to any object. To keep this functionality with the wrapper
        # BaseFromRaw you can add annotations in a nested dict.

        # `_generate_minimal_annotations()` must be called to generate the nested
        # dict of annotations/array_annotations
        self._generate_minimal_annotations()
        # this pprint lines really help with understanding the nested (and sometimes complicated) dict
        # from pprint import pprint
        # pprint(self.raw_annotations)

        # Until here all mandatory operations for setting up a rawio are implemented.
        # The following lines provide additional, recommended annotations for the
        # final neo objects.
        for block_index in range(2):
            bl_ann = self.raw_annotations["blocks"][block_index]
            bl_ann["name"] = f"Block #{block_index}"
            bl_ann["block_extra_info"] = f"This is the block {block_index}"
            for seg_index in range([2, 3][block_index]):
                seg_ann = bl_ann["segments"][seg_index]
                seg_ann["name"] = f"Seg #{seg_index} Block #{block_index}"
                seg_ann["seg_extra_info"] = f"This is the seg {seg_index} of block {block_index}"
                for c in range(2):
                    sig_an = seg_ann["signals"][c]["nickname"] = f"This stream {c} is from a subdevice"
                    # add some array annotations (8 channels)
                    sig_an = seg_ann["signals"][c]["__array_annotations__"]["impedance"] = np.random.rand(8) * 10000
                for c in range(3):
                    spiketrain_an = seg_ann["spikes"][c]
                    spiketrain_an["quality"] = "Good!!"
                    # add some array annotations
                    num_spikes = self.spike_count(block_index, seg_index, c)
                    spiketrain_an["__array_annotations__"]["amplitudes"] = np.random.randn(num_spikes)

                for c in range(2):
                    event_an = seg_ann["events"][c]
                    if c == 0:
                        event_an["nickname"] = "Miss Event 0"
                        # add some array annotations
                        num_ev = self.event_count(block_index, seg_index, c)
                        event_an["__array_annotations__"]["button"] = ["A"] * num_ev
                    elif c == 1:
                        event_an["nickname"] = "MrEpoch 1"

    def _segment_t_start(self, block_index: int, seg_index: int):
        # this must return a float scaled in seconds
        # this t_start will be shared by all objects in the segment
        # except AnalogSignal
        all_starts = [[0.0, 15.0], [0.0, 20.0, 60.0]]
        return all_starts[block_index][seg_index]

    def _segment_t_stop(self, block_index: int, seg_index: int):
        # this must return a float scaled in seconds
        all_stops = [[10.0, 25.0], [10.0, 30.0, 70.0]]
        return all_stops[block_index][seg_index]

    def _get_signal_size(self, block_index: int, seg_index: int, stream_index: int):
        # We generate fake data in which the two stream signals have the same shape
        # across all segments (10.0 seconds)
        # This is not the case for real data, instead you should return the signal
        # size depending on the block_index and segment_index
        # this must return an int = the number of samples

        # Note that channel_indexes can be ignored for most cases
        # except for the case of several sampling rates.
        return 100000

    def _get_signal_t_start(self, block_index: int, seg_index: int, stream_index: int):
        # This give the t_start of a signal.
        # Very often this is equal to _segment_t_start but not
        # always.
        # this must return a float scaled in seconds

        # Note that channel_indexes can be ignored for most cases
        # except for the case of several sampling rates.

        # Here this is the same.
        # this is not always the case
        return self._segment_t_start(block_index, seg_index)

    def _get_analogsignal_chunk(
        self,
        block_index: int,
        seg_index: int,
        i_start: int | None,
        i_stop: int | None,
        stream_index: int,
        channel_indexes: np.ndarray | list | slice | None,
    ):
        # this must return a signal chunk in a signal stream
        # limited with i_start/i_stop (can be None)
        # channel_indexes can be None (=all channel in the stream) or a list or numpy.array
        # This must return a numpy array 2D (even with one channel).
        # This must return the original dtype. No conversion here.
        # This must be as fast as possible.
        # To speed up this call all preparatory calculations should be implemented
        # in _parse_header().

        # Here we are lucky:  our signals are always zeros!!
        # it is not always the case :)
        # internally signals are int16
        # conversion to real units is done with self.header['signal_channels']

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = 100000

        if i_start < 0 or i_stop > 100000:
            # some checks
            raise IndexError("I don't like your jokes")

        if channel_indexes is None:
            nb_chan = 8
        elif isinstance(channel_indexes, slice):
            channel_indexes = np.arange(8, dtype="int")[channel_indexes]
            nb_chan = len(channel_indexes)
        else:
            channel_indexes = np.asarray(channel_indexes)
            if any(channel_indexes < 0):
                raise IndexError("bad boy")
            if any(channel_indexes >= 8):
                raise IndexError("big bad wolf")
            nb_chan = len(channel_indexes)

        raw_signals = np.zeros((i_stop - i_start, nb_chan), dtype="int16")
        return raw_signals

    def _spike_count(self, block_index: int, seg_index: int, spike_channel_index: int):
        # Must return the nb of spikes for given (block_index, seg_index, spike_channel_index)
        # we are lucky:  our units have all the same nb of spikes!!
        # it is not always the case
        nb_spikes = 20
        return nb_spikes

    def _get_spike_timestamps(
        self, block_index: int, seg_index: int, spike_channel_index: int, t_start: float | None, t_stop: float | None
    ):
        # In our IO, timestamp are internally coded 'int64' and they
        # represent the index of the signals 10kHz
        # we are lucky: spikes have the same discharge in all segments!!
        # incredible neuron!! This is not always the case

        # the same clip t_start/t_start must be used in _spike_raw_waveforms()

        ts_start = self._segment_t_start(block_index, seg_index) * 10000

        spike_timestamps = np.arange(0, 10000, 500) + ts_start

        if t_start is not None or t_stop is not None:
            # restrict spikes to given limits (in seconds)
            lim0 = int(t_start * 10000)
            lim1 = int(t_stop * 10000)
            mask = (spike_timestamps >= lim0) & (spike_timestamps <= lim1)
            spike_timestamps = spike_timestamps[mask]

        return spike_timestamps

    def _rescale_spike_timestamp(self, spike_timestamps: np.ndarray, dtype: np.dtype):
        # must rescale to seconds, a particular spike_timestamps
        # with a fixed dtype so the user can choose the precision they want.
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= 10000.0  # because 10kHz
        return spike_times

    def _get_spike_raw_waveforms(
        self, block_index: int, seg_index: int, spike_channel_index: int, t_start: float | None, t_stop: float | None
    ):
        # this must return a 3D numpy array (nb_spike, nb_channel, nb_sample)
        # in the original dtype
        # this must be as fast as possible.
        # the same clip t_start/t_start must be used in _spike_timestamps()

        # If there there is no waveform supported in the
        # IO them _spike_raw_waveforms must return None

        # In our IO waveforms come from all channels
        # they are int16
        # conversion to real units is done with self.header['spike_channels']
        # Here, we have a realistic case: all waveforms are only noise.
        # it is not always the case
        # we get 20 spikes with a sweep of 50 (5ms)

        # trick to get how many spike in the slice
        ts = self._get_spike_timestamps(block_index, seg_index, spike_channel_index, t_start, t_stop)
        nb_spike = ts.size

        np.random.seed(2205)  # a magic number (my birthday)
        waveforms = np.random.randint(low=-(2**4), high=2**4, size=nb_spike * 50, dtype="int16")
        waveforms = waveforms.reshape(nb_spike, 1, 50)
        return waveforms

    def _event_count(self, block_index: int, seg_index: int, event_channel_index: int):
        # event and spike are very similar
        # we have 2 event channels
        if event_channel_index == 0:
            # event channel
            return 6
        elif event_channel_index == 1:
            # epoch channel
            return 10

    def _get_event_timestamps(
        self, block_index: int, seg_index: int, event_channel_index: int, t_start: float | None, t_stop: float | None
    ):
        # the main difference between spike channel and event channel
        # is that for event channels we have 3D numpy array (timestamp, durations, labels) where
        # durations must be None for 'event'
        # label must a dtype ='U'

        # in our IO events are directly coded in seconds
        seg_t_start = self._segment_t_start(block_index, seg_index)
        if event_channel_index == 0:
            timestamp = np.arange(0, 6, dtype="float64") + seg_t_start
            durations = None
            labels = np.array(["trigger_a", "trigger_b"] * 3, dtype="U12")
        elif event_channel_index == 1:
            timestamp = np.arange(0, 10, dtype="float64") + 0.5 + seg_t_start
            durations = np.ones((10), dtype="float64") * 0.25
            labels = np.array(["zoneX"] * 5 + ["zoneZ"] * 5, dtype="U12")

        if t_start is not None:
            keep = timestamp >= t_start
            timestamp, labels = timestamp[keep], labels[keep]
            if durations is not None:
                durations = durations[keep]

        if t_stop is not None:
            keep = timestamp <= t_stop
            timestamp, labels = timestamp[keep], labels[keep]
            if durations is not None:
                durations = durations[keep]

        return timestamp, durations, labels

    def _rescale_event_timestamp(self, event_timestamps: np.ndarray, dtype: np.dtype, event_channel_index: int):
        # must rescale to seconds for a particular event_timestamps
        # with a fixed dtype so the user can choose the precision they want.

        # really easy here because in our case it is already in seconds
        event_times = event_timestamps.astype(dtype)
        return event_times

    def _rescale_epoch_duration(self, raw_duration: np.ndarray, dtype: np.dtype, event_channel_index: int):
        # really easy here because in our case it is already in seconds
        durations = raw_duration.astype(dtype)
        return durations



"""
Class for reading data from NeuroExplorer (.nex)

Note:
  * NeuroExplorer have introduced a new .nex5 file format
    with 64 timestamps. This is NOT implemented here.
    If someone have some file in that new format we could also
    integrate it in neo
  * NeuroExplorer now provide there own python class for
    reading/writing nex and nex5. This could be useful
    for testing this class.

Porting NeuroExplorerIO to NeuroExplorerRawIO have some
limitation because in neuro explorer signals can have different sampling
rate and shape. So NeuroExplorerRawIO can read only one channel
at once.

Documentation for dev :
http://www.neuroexplorer.com/downloadspage/


Author: Samuel Garcia, luc estebanez, mark hollenbeck

"""

from collections import OrderedDict

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)

from neo.core import NeoReadWriteError


class NeuroExplorerRawIO(BaseRawIO):
    extensions = ["nex"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        BaseRawIO.__init__(self)
        self.filename = filename

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        with open(self.filename, "rb") as fid:
            self.global_header = read_as_dict(fid, GlobalHeader, offset=0)
            offset = 544
            self._entity_headers = []
            for i in range(self.global_header["nvar"]):
                self._entity_headers.append(read_as_dict(fid, EntityHeader, offset=offset + i * 208))

        self._memmap = np.memmap(self.filename, dtype="u1", mode="r")

        self._sig_lengths = []
        self._sig_t_starts = []
        sig_channels = []
        spike_channels = []
        event_channels = []
        for i in range(self.global_header["nvar"]):
            entity_header = self._entity_headers[i]
            name = entity_header["name"]
            _id = str(i)
            if entity_header["type"] == 0:  # Unit
                spike_channels.append((name, _id, "", 0, 0, 0, 0))

            elif entity_header["type"] == 1:  # Event
                event_channels.append((name, _id, "event"))

            elif entity_header["type"] == 2:  # interval = Epoch
                event_channels.append((name, _id, "epoch"))

            elif entity_header["type"] == 3:  # spiketrain and waveforms
                wf_units = "mV"
                wf_gain = entity_header["ADtoMV"]
                wf_offset = entity_header["MVOffset"]
                wf_left_sweep = 0
                wf_sampling_rate = entity_header["WFrequency"]
                spike_channels.append((name, _id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate))

            elif entity_header["type"] == 4:
                # popvectors
                pass

            if entity_header["type"] == 5:  # Signals
                units = "mV"
                sampling_rate = entity_header["WFrequency"]
                dtype = "int16"
                gain = entity_header["ADtoMV"]
                offset = entity_header["MVOffset"]
                stream_id = str(_id)
                sig_channels.append((name, _id, sampling_rate, dtype, units, gain, offset, stream_id))
                self._sig_lengths.append(entity_header["NPointsWave"])
                # sig t_start is the first timestamp if datablock
                offset = entity_header["offset"]
                timestamps0 = self._memmap[offset : offset + 4].view("int32")
                t_start = timestamps0[0] / self.global_header["freq"]
                self._sig_t_starts.append(t_start)

            elif entity_header["type"] == 6:  # Markers
                event_channels.append((name, _id, "event"))

        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # each signal channel has different groups that force reading
        # them one by one
        sig_channels["stream_id"] = np.arange(sig_channels.size).astype("U")
        signal_streams = np.zeros(sig_channels.size, dtype=_signal_stream_dtype)
        signal_streams["name"] = sig_channels["name"]
        signal_streams["id"] = sig_channels["stream_id"]

        # fill into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = sig_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # Annotations
        self._generate_minimal_annotations()
        bl_annotations = self.raw_annotations["blocks"][0]
        seg_annotations = bl_annotations["segments"][0]
        for d in (bl_annotations, seg_annotations):
            d["neuroexplorer_version"] = self.global_header["version"]
            d["comment"] = self.global_header["comment"]

    def _segment_t_start(self, block_index, seg_index):
        t_start = self.global_header["tbeg"] / self.global_header["freq"]
        return t_start

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self.global_header["tend"] / self.global_header["freq"]
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        return self._sig_lengths[stream_index]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return self._sig_t_starts[stream_index]

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        channel_index = stream_index
        entity_index = int(self.header["signal_channels"][channel_index]["id"])
        entity_header = self._entity_headers[entity_index]
        n = entity_header["n"]
        nb_sample = entity_header["NPointsWave"]
        # offset = entity_header['offset']
        # timestamps = self._memmap[offset:offset+n*4].view('int32')
        # offset2 = entity_header['offset'] + n*4
        # fragment_starts = self._memmap[offset2:offset2+n*4].view('int32')
        offset3 = entity_header["offset"] + n * 4 + n * 4
        raw_signal = self._memmap[offset3 : offset3 + nb_sample * 2].view("int16")
        raw_signal = raw_signal[slice(i_start, i_stop), None]  # 2D for compliance
        return raw_signal

    def _spike_count(self, block_index, seg_index, unit_index):
        entity_index = int(self.header["spike_channels"][unit_index]["id"])
        entity_header = self._entity_headers[entity_index]
        nb_spike = entity_header["n"]
        return nb_spike

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        entity_index = int(self.header["spike_channels"][unit_index]["id"])
        entity_header = self._entity_headers[entity_index]
        n = entity_header["n"]
        offset = entity_header["offset"]
        timestamps = self._memmap[offset : offset + n * 4].view("int32")

        if t_start is not None:
            keep = timestamps >= int(t_start * self.global_header["freq"])
            timestamps = timestamps[keep]
        if t_stop is not None:
            keep = timestamps <= int(t_stop * self.global_header["freq"])
            timestamps = timestamps[keep]

        return timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= self.global_header["freq"]
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        entity_index = int(self.header["spike_channels"][unit_index]["id"])
        entity_header = self._entity_headers[entity_index]
        if entity_header["type"] == 0:
            return None
        if entity_header["type"] != 3:
            raise NeoReadWriteError(f"Neo requires the entity_header['type'] to be 3 not {entity_header['type']}")

        n = entity_header["n"]
        width = entity_header["NPointsWave"]
        offset = entity_header["offset"] + n * 2
        waveforms = self._memmap[offset : offset + n * 2 * width].view("int16")
        waveforms = waveforms.reshape(n, 1, width)

        return waveforms

    def _event_count(self, block_index, seg_index, event_channel_index):
        entity_index = int(self.header["event_channels"][event_channel_index]["id"])
        entity_header = self._entity_headers[entity_index]
        nb_event = entity_header["n"]
        return nb_event

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        entity_index = int(self.header["event_channels"][event_channel_index]["id"])
        entity_header = self._entity_headers[entity_index]

        n = entity_header["n"]
        offset = entity_header["offset"]
        timestamps = self._memmap[offset : offset + n * 4].view("int32")

        if t_start is None:
            i_start = None
        else:
            i_start = np.searchsorted(timestamps, int(t_start * self.global_header["freq"]))
        if t_stop is None:
            i_stop = None
        else:
            i_stop = np.searchsorted(timestamps, int(t_stop * self.global_header["freq"]))
        keep = slice(i_start, i_stop)

        timestamps = timestamps[keep]

        if entity_header["type"] == 1:  # Event
            durations = None
            labels = np.array([""] * timestamps.size, dtype="U")
        elif entity_header["type"] == 2:  # Epoch
            offset2 = offset + n * 4
            stop_timestamps = self._memmap[offset2 : offset2 + n * 4].view("int32")
            durations = stop_timestamps[keep] - timestamps
            labels = np.array([""] * timestamps.size, dtype="U")
        elif entity_header["type"] == 6:  # Marker
            durations = None
            offset2 = offset + n * 4 + 64
            s = entity_header["MarkerLength"]
            labels = self._memmap[offset2 : offset2 + s * n].view("S" + str(s))
            labels = labels[keep].astype("U")

        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = event_timestamps.astype(dtype)
        event_times /= self.global_header["freq"]
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        durations = raw_duration.astype(dtype)
        durations /= self.global_header["freq"]
        return durations


def read_as_dict(fid, dtype, offset=None):
    """
    Given a file descriptor
    and a numpy.dtype of the binary struct return a dict.
    Make conversion for strings.
    """
    if offset is not None:
        fid.seek(offset)
    dt = np.dtype(dtype)
    h = np.frombuffer(fid.read(dt.itemsize), dt)[0]
    info = OrderedDict()
    for k in dt.names:
        v = h[k]

        if dt[k].kind == "S":
            v = v.replace(b"\x00", b"")
            v = v.decode("utf8")

        info[k] = v
    return info


GlobalHeader = [
    ("signature", "S4"),
    ("version", "int32"),
    ("comment", "S256"),
    ("freq", "float64"),
    ("tbeg", "int32"),
    ("tend", "int32"),
    ("nvar", "int32"),
]

EntityHeader = [
    ("type", "int32"),
    ("varVersion", "int32"),
    ("name", "S64"),
    ("offset", "int32"),
    ("n", "int32"),
    ("WireNumber", "int32"),
    ("UnitNumber", "int32"),
    ("Gain", "int32"),
    ("Filter", "int32"),
    ("XPos", "float64"),
    ("YPos", "float64"),
    ("WFrequency", "float64"),
    ("ADtoMV", "float64"),
    ("NPointsWave", "int32"),
    ("NMarkers", "int32"),
    ("MarkerLength", "int32"),
    ("MVOffset", "float64"),
    ("dummy", "S60"),
]

MarkerHeader = [
    ("type", "int32"),
    ("varVersion", "int32"),
    ("name", "S64"),
    ("offset", "int32"),
    ("n", "int32"),
    ("WireNumber", "int32"),
    ("UnitNumber", "int32"),
    ("Gain", "int32"),
    ("Filter", "int32"),
]



import datetime
from packaging.version import Version
import os
import re
from collections import OrderedDict


class NlxHeader(OrderedDict):
    """
    Representation of basic information in all 16 kbytes Neuralynx file headers,
    including dates opened and closed if given.
    """

    HEADER_SIZE = 2**14  # Neuralynx files have a txt header of 16kB

    # helper function to interpret boolean keys
    def _to_bool(txt):
        if txt == "True":
            return True
        elif txt == "False":
            return False
        else:
            raise Exception("Can not convert %s to bool" % txt)

    # keys that may be present in header which we parse
    txt_header_keys = [
        ("AcqEntName", "channel_names", None),  # used
        ("FileType", "", None),
        ("FileVersion", "", None),
        ("RecordSize", "", None),
        ("HardwareSubSystemName", "", None),
        ("HardwareSubSystemType", "", None),
        ("SamplingFrequency", "sampling_rate", float),  # used
        ("ADMaxValue", "", None),
        ("ADBitVolts", "bit_to_microVolt", None),  # used
        ("NumADChannels", "", None),
        ("ADChannel", "channel_ids", None),  # used
        ("InputRange", "", None),
        ("InputInverted", "input_inverted", _to_bool),  # used
        ("DSPLowCutFilterEnabled", "", None),
        ("DspLowCutFrequency", "", None),
        ("DspLowCutNumTaps", "", None),
        ("DspLowCutFilterType", "", None),
        ("DSPHighCutFilterEnabled", "", None),
        ("DspHighCutFrequency", "", None),
        ("DspHighCutNumTaps", "", None),
        ("DspHighCutFilterType", "", None),
        ("DspDelayCompensation", "", None),
        ("DspFilterDelay_µs", "", None),
        ("DisabledSubChannels", "", None),
        ("WaveformLength", "", int),
        ("AlignmentPt", "", None),
        ("ThreshVal", "", None),
        ("MinRetriggerSamples", "", None),
        ("SpikeRetriggerTime", "", None),
        ("DualThresholding", "", None),
        (r"Feature \w+ \d+", "", None),
        ("SessionUUID", "", None),
        ("FileUUID", "", None),
        ("CheetahRev", "", None),  # only for older versions of Cheetah
        ("ProbeName", "", None),
        ("OriginalFileName", "", None),
        ("TimeCreated", "", None),
        ("TimeClosed", "", None),
        ("ApplicationName", "", None),  # also include version number when present
        ("AcquisitionSystem", "", None),
        ("ReferenceChannel", "", None),
        ("NLX_Base_Class_Type", "", None),  # in version 4 and earlier versions of Cheetah
    ]

    # Filename and datetime may appear in header lines starting with # at
    # beginning of header or in later versions as a property. The exact format
    # used depends on the application name and its version as well as the
    # -FileVersion property.
    #
    # There are 4 styles understood by this code and the patterns used for parsing
    # the items within each are stored in a dictionary. Each dictionary is then
    # stored in main dictionary keyed by an abbreviation for the style.
    header_pattern_dicts = {
        # BML
        "bml": dict(
            datetime1_regex=r"## Time Opened: \(m/d/y\): (?P<date>\S+)" r"  At Time: (?P<time>\S+)",
            filename_regex=r"## File Name: (?P<filename>\S+)",
            datetimeformat="%m/%d/%y %H:%M:%S.%f",
        ),
        # Cheetah after version 1 and before version 5
        "bv5": dict(
            datetime1_regex=r"## Time Opened: \(m/d/y\): (?P<date>\S+)" r"  At Time: (?P<time>\S+)",
            filename_regex=r"## File Name: (?P<filename>\S+)",
            datetimeformat="%m/%d/%Y %H:%M:%S.%f",
        ),
        # Cheetah version 5.4.0
        "v5.4.0": dict(
            datetime1_regex=r"## Time Opened \(m/d/y\): (?P<date>\S+)" r"  At Time: (?P<time>\S+)",
            datetime2_regex=r"## Time Closed \(m/d/y\): (?P<date>\S+)" r"  At Time: (?P<time>\S+)",
            filename_regex=r"## File Name: (?P<filename>\S+)",
            datetimeformat="%m/%d/%Y %H:%M:%S.%f",
        ),
        # Cheetah version 5.6.0, some range of versions in between
        "v5.6.0": dict(
            datetime1_regex=r"## Time Opened: \(m/d/y\): (?P<date>\S+)" r" At Time: (?P<time>\S+)",
            filename_regex=r"## File Name: (?P<filename>\S+)",
            datetimeformat="%m/%d/%Y %H:%M:%S.%f",
        ),
        # Cheetah version 5 before and including v 5.6.4 as well as version 1
        "bv5.6.4": dict(
            datetime1_regex=r"## Time Opened \(m/d/y\): (?P<date>\S+)" r"  \(h:m:s\.ms\) (?P<time>\S+)",
            datetime2_regex=r"## Time Closed \(m/d/y\): (?P<date>\S+)" r"  \(h:m:s\.ms\) (?P<time>\S+)",
            filename_regex=r"## File Name (?P<filename>\S+)",
            datetimeformat="%m/%d/%Y %H:%M:%S.%f",
        ),
        "neuraview2": dict(
            datetime1_regex=r"## Date Opened: \(mm/dd/yyy\): (?P<date>\S+)" r" At Time: (?P<time>\S+)",
            datetime2_regex=r"## Date Closed: \(mm/dd/yyy\): (?P<date>\S+)" r" At Time: (?P<time>\S+)",
            filename_regex=r"## File Name: (?P<filename>\S+)",
            datetimeformat="%m/%d/%Y %H:%M:%S",
        ),
        "peg": dict(
            datetime1_regex=r"-TimeCreated (?P<date>\S+) (?P<time>\S+)",
            datetime2_regex=r"-TimeClosed (?P<date>\S+) (?P<time>\S+)",
            filename_regex=r'-OriginalFileName "?(?P<filename>\S+)"?',
            datetimeformat=r"%Y/%m/%d %H:%M:%S",
            datetime2format=r"%Y/%m/%d %H:%M:%S.%f",
        ),
        # Cheetah after v 5.6.4 and default for others such as Pegasus
        "def": dict(
            datetime1_regex=r"-TimeCreated (?P<date>\S+) (?P<time>\S+)",
            datetime2_regex=r"-TimeClosed (?P<date>\S+) (?P<time>\S+)",
            filename_regex=r'-OriginalFileName "?(?P<filename>\S+)"?',
            datetimeformat="%Y/%m/%d %H:%M:%S",
        ),
    }

    def __init__(self, filename, props_only=False):
        """
        Factory function to build NlxHeader for a given file.

        :param filename: name of Neuralynx file
        :param props_only: if true, will not try and read time and date or check start
        """
        super(OrderedDict, self).__init__()
        with open(filename, "rb") as f:
            txt_header = f.read(NlxHeader.HEADER_SIZE)
        txt_header = txt_header.strip(b"\x00").decode("latin-1")

        # must start with 8 # characters
        if not props_only and not txt_header.startswith("########"):
            ValueError("Neuralynx files must start with 8 # characters.")

        self.read_properties(filename, txt_header)

        if not props_only:
            self.readTimeDate(txt_header)

    @staticmethod
    def build_with_properties_only(filename):
        """
        Builds a version of the header without time and date or other validity checking.

        Intended mostly for utilities but may also be useful for some recalcitrant header formats.

        :param filename: name of Neuralynx file.
        :return: NlxHeader with properties from header text
        """
        res = OrderedDict()

    def read_properties(self, filename, txt_header):
        """
        Read properties from header and place in OrderedDictionary which this object is.
        :param filename: name of ncs file, used for extracting channel number
        :param txt_header: header text
        """
        # find keys
        for k1, k2, type_ in NlxHeader.txt_header_keys:
            pattern = r"-(?P<name>" + k1 + r")\s+(?P<value>[\S ]*)"
            matches = re.findall(pattern, txt_header)
            for match in matches:
                if k2 == "":
                    name = match[0]
                else:
                    name = k2
                value = match[1].rstrip(" ")
                if type_ is not None:
                    value = type_(value)
                self[name] = value
        # if channel_ids or s not in self then the filename is used
        name = os.path.splitext(os.path.basename(filename))[0]
        # convert channel ids
        if "channel_ids" in self:
            chid_entries = re.findall(r"\S+", self["channel_ids"])
            self["channel_ids"] = [int(c) for c in chid_entries]
        else:
            self["channel_ids"] = ["unknown"]
        # convert channel names
        if "channel_names" in self:
            name_entries = re.findall(r"\S+", self["channel_names"])
            if len(name_entries) == 1:
                self["channel_names"] = name_entries * len(self["channel_ids"])
            assert len(self["channel_names"]) == len(
                self["channel_ids"]
            ), "Number of channel ids does not match channel names."
        else:
            self["channel_names"] = ["unknown"] * len(self["channel_ids"])
        # version and application name
        # older Cheetah versions with CheetahRev property
        if "CheetahRev" in self:
            assert "ApplicationName" not in self
            self["ApplicationName"] = "Cheetah"
            app_version = self["CheetahRev"]
        # new file version 3.4 does not contain CheetahRev property, but ApplicationName instead
        elif "ApplicationName" in self:
            pattern = r'(\S*) "([\S ]*)"'
            match = re.findall(pattern, self["ApplicationName"])
            assert len(match) == 1, "impossible to find application name and version"
            self["ApplicationName"], app_version = match[0]
        # BML Ncs file contain neither property, but 'NLX_Base_Class_Type'
        elif "NLX_Base_Class_Type" in txt_header:
            self["ApplicationName"] = "BML"
            app_version = "2.0"
        # Neuraview Ncs file contained neither property nor
        # NLX_Base_Class_Type information
        else:
            self["ApplicationName"] = "Neuraview"
            app_version = "2"
        if " Development" in app_version:
            app_version = app_version.replace(" Development", ".dev0")
        self["ApplicationVersion"] = Version(app_version)
        # convert bit_to_microvolt
        if "bit_to_microVolt" in self:
            btm_entries = re.findall(r"\S+", self["bit_to_microVolt"])
            if len(btm_entries) == 1:
                btm_entries = btm_entries * len(self["channel_ids"])
            self["bit_to_microVolt"] = [float(e) * 1e6 for e in btm_entries]
            assert len(self["bit_to_microVolt"]) == len(
                self["channel_ids"]
            ), "Number of channel ids does not match bit_to_microVolt conversion factors."
        if "InputRange" in self:
            ir_entries = re.findall(r"\w+", self["InputRange"])
            if len(ir_entries) == 1:
                self["InputRange"] = [int(ir_entries[0])] * len(chid_entries)
            else:
                self["InputRange"] = [int(e) for e in ir_entries]
            assert len(self["InputRange"]) == len(
                chid_entries
            ), "Number of channel ids does not match input range values."

    def readTimeDate(self, txt_header):
        """
        Read time and date from text of header appropriate for app name and version

        :TODO: this works for current examples but is not likely actually related
        to app version in this manner.
        """
        an = self["ApplicationName"]
        if an == "Cheetah":
            av = self["ApplicationVersion"]
            if av <= Version("2"):  # version 1 uses same as older versions
                hpd = NlxHeader.header_pattern_dicts["bv5.6.4"]
            elif av < Version("5"):
                hpd = NlxHeader.header_pattern_dicts["bv5"]
            elif av <= Version("5.4.0"):
                hpd = NlxHeader.header_pattern_dicts["v5.4.0"]
            elif av == Version("5.6.0"):
                hpd = NlxHeader.header_pattern_dicts["v5.6.0"]
            elif av <= Version("5.6.4"):
                hpd = NlxHeader.header_pattern_dicts["bv5.6.4"]
            else:
                hpd = NlxHeader.header_pattern_dicts["def"]
        elif an == "BML":
            hpd = NlxHeader.header_pattern_dicts["bml"]
            av = Version("2")
        elif an == "Neuraview":
            hpd = NlxHeader.header_pattern_dicts["neuraview2"]
            av = Version("2")
        elif an == "Pegasus":
            hpd = NlxHeader.header_pattern_dicts["peg"]
            av = Version("2")
        else:
            an = "Unknown"
            av = "NA"
            hpd = NlxHeader.header_pattern_dicts["def"]
        # opening time
        sr = re.search(hpd["datetime1_regex"], txt_header)
        if not sr:
            raise IOError(
                f"No matching header open date/time for application {an} " + f"version {av}. Please contact developers."
            )
        else:
            dt1 = sr.groupdict()
            try:  # allow two possible formats for date and time
                self["recording_opened"] = datetime.datetime.strptime(
                    dt1["date"] + " " + dt1["time"], hpd["datetimeformat"]
                )
            except:
                try:
                    self["recording_opened"] = datetime.datetime.strptime(
                        dt1["date"] + " " + dt1["time"], hpd["datetime2format"]
                    )
                except:
                    self["recording_opened"] = None

        # close time, if available
        if "datetime2_regex" in hpd:
            sr = re.search(hpd["datetime2_regex"], txt_header)
            if not sr:
                raise IOError(
                    f"No matching header close date/time for application {an} "
                    + f"version {av}. Please contact developers."
                )
            else:
                try:
                    dt2 = sr.groupdict()
                    self["recording_closed"] = datetime.datetime.strptime(
                        dt2["date"] + " " + dt2["time"], hpd["datetimeformat"]
                    )
                except:
                    self["recording_closed"] = None

    def type_of_recording(self):
        """
        Determines type of recording in Ncs file with this header.

        RETURN:
            one of 'PRE4','BML','DIGITALLYNX','DIGITALLYNXSX','UNKNOWN'
        """

        if "NLX_Base_Class_Type" in self:

            # older style standard neuralynx acquisition with rounded sampling frequency
            if self["NLX_Base_Class_Type"] == "CscAcqEnt":
                return "PRE4"

            # BML style with fractional frequency and microsPerSamp
            elif self["NLX_Base_Class_Type"] == "BmlAcq":
                return "BML"

            else:
                return "UNKNOWN"

        elif "HardwareSubSystemType" in self:

            # DigitalLynx
            if self["HardwareSubSystemType"] == "DigitalLynx":
                return "DIGITALLYNX"

            # DigitalLynxSX
            elif self["HardwareSubSystemType"] == "DigitalLynxSX":
                return "DIGITALLYNXSX"

            # Cheetah64
            elif self["HardwareSubSystemType"] == "Cheetah64":
                return "CHEETAH64"

            # RawDataFile
            elif self["HardwareSubSystemType"] == "RawDataFile":
                return "RAWDATAFILE"

            else:
                return "UNKNOWN"

        elif "FileType" in self:

            if "FileVersion" in self and self["FileVersion"] in ["3.2", "3.3", "3.4"]:
                return self["AcquisitionSystem"].split()[1].upper()

            else:
                return "CHEETAH560"  # only known case of FileType without FileVersion

        else:
            return "UNKNOWN"



"""
Here a list of proxy object that can be used when lazy=True at neo.io level.

This idea is to be able to postpone that real in memory loading
for objects that contains big data (AnalogSIgnal, SpikeTrain, Event, Epoch).

The implementation rely on neo.rawio, so it will available only for neo.io that
ineherits neo.rawio.

"""

import logging

import numpy as np
import quantities as pq


from neo.core.baseneo import BaseNeo
from neo.core import AnalogSignal, Epoch, Event, SpikeTrain
from neo.core.dataobject import ArrayDict


logger = logging.getLogger("Neo")


class BaseProxy(BaseNeo):
    def __init__(self, array_annotations=None, **annotations):
        # this for py27 str vs py3 str in neo attributes ompatibility
        annotations = check_annotations(annotations)
        if "file_origin" not in annotations:
            # the str is to make compatible with neo_py27 where attribute
            # used to be str so raw bytes
            annotations["file_origin"] = str(self._rawio.source_name())

        if array_annotations is None:
            array_annotations = {}
        for k, v in array_annotations.items():
            array_annotations[k] = np.asarray(v)

        # clean array annotations that are not 1D
        # TODO remove this once multi-dimensional array_annotations are possible
        array_annotations = {k: v for k, v in array_annotations.items() if v.ndim == 1}

        # this mock the array annotations to avoid inherits DataObject
        self.array_annotations = ArrayDict(self.shape[-1])
        self.array_annotations.update(array_annotations)

        BaseNeo.__init__(self, **annotations)

    def load(self, time_slice=None, **kwargs):
        # should be implemented by subclass
        raise NotImplementedError

    def time_slice(self, t_start, t_stop):
        """
        Load the proxy object within the specified time range. Has the same
        call signature as AnalogSignal.time_slice, Epoch.time_slice, etc.
        """
        return self.load(time_slice=(t_start, t_stop))


class AnalogSignalProxy(BaseProxy):
    """
    This object mimic AnalogSignal except that it does not
    have the signals array itself. All attributes and annotations are here.

    The goal is to postpone the loading of data into memory
    when reading a file with the new lazy load system based
    on neo.rawio.

    This object must not be constructed directly but is given
    neo.io when lazy=True instead of a true AnalogSignal.

    The AnalogSignalProxy is able to load:
      * only a slice of time
      * only a subset of channels
      * have an internal raw magnitude identic to the file (int16) with
        a pq.CompoundUnit().

    Usage:
    >>> proxy_anasig = AnalogSignalProxy(rawio=self.reader,
                                         global_channel_indexes=None,
                                         block_index=0,
                                         seg_index=0)
    >>> anasig = proxy_anasig.load()
    >>> slice_of_anasig = proxy_anasig.load(time_slice=(1.*pq.s, 2.*pq.s))
    >>> some_channel_of_anasig = proxy_anasig.load(channel_indexes=[0,5,10])

    """

    _parent_objects = ("Segment",)
    _necessary_attrs = (("sampling_rate", pq.Quantity, 0), ("t_start", pq.Quantity, 0))
    _recommended_attrs = BaseNeo._recommended_attrs
    proxy_for = AnalogSignal

    def __init__(self, rawio=None, stream_index=None, inner_stream_channels=None, block_index=0, seg_index=0):
        # stream_index:  indicate the stream stream_id can be retreive easily
        # inner_stream_channels: are channel index inside the stream None means all channels
        # if inner_stream_channels is not None:
        #     * then this is a "substream"
        #     * handle the case where channels have different units inside a stream
        #     * is related to BaseFromRaw.get_sub_signal_streams()

        self._rawio = rawio
        self._block_index = block_index
        self._seg_index = seg_index
        self._stream_index = stream_index
        if inner_stream_channels is None:
            inner_stream_channels = slice(inner_stream_channels)
        self._inner_stream_channels = inner_stream_channels

        signal_streams = self._rawio.header["signal_streams"]
        stream_id = signal_streams[stream_index]["id"]
        signal_channels = self._rawio.header["signal_channels"]
        (global_inds,) = np.nonzero(signal_channels["stream_id"] == stream_id)
        self._nb_total_chann_in_stream = global_inds.size
        self._global_channel_indexes = global_inds[inner_stream_channels]
        self._nb_chan = self._global_channel_indexes.size

        sig_chans = signal_channels[self._global_channel_indexes]

        if np.unique(sig_chans["units"]).size != 1:
            raise ValueError("Channel do not have same units")
        if np.unique(sig_chans["dtype"]).size != 1:
            raise TypeError("Channel do not have same dtype")
        if np.unique(sig_chans["sampling_rate"]).size != 1:
            raise ValueError("Channel do not have same sampling_rate")

        self.units = ensure_signal_units(sig_chans["units"][0])
        self.dtype = sig_chans["dtype"][0]
        self.sampling_rate = sig_chans["sampling_rate"][0] * pq.Hz
        self.sampling_period = 1.0 / self.sampling_rate
        sigs_size = self._rawio.get_signal_size(block_index=block_index, seg_index=seg_index, stream_index=stream_index)
        self.shape = (sigs_size, self._nb_chan)
        self.t_start = self._rawio.get_signal_t_start(block_index, seg_index, stream_index) * pq.s

        # magnitude_mode='raw' is supported only if all offset=0
        # and all gain are the same
        support_raw_magnitude = np.all(sig_chans["gain"] == sig_chans["gain"][0]) and np.all(sig_chans["offset"] == 0.0)

        if support_raw_magnitude:
            str_units = ensure_signal_units(sig_chans["units"][0]).units.dimensionality.string
            gain0 = sig_chans["gain"][0]
            self._raw_units = pq.CompoundUnit(f"{gain0}*{str_units}")
        else:
            self._raw_units = None

        # retrieve annotations and array annotations
        seg_ann = self._rawio.raw_annotations["blocks"][block_index]["segments"][seg_index]
        annotations = seg_ann["signals"][stream_index].copy()
        array_annotations = annotations.pop("__array_annotations__")
        array_annotations = {k: v[inner_stream_channels] for k, v in array_annotations.items()}

        BaseProxy.__init__(self, array_annotations=array_annotations, **annotations)

    @property
    def duration(self):
        """Signal duration"""
        return self.shape[0] / self.sampling_rate

    @property
    def t_stop(self):
        """Time when signal ends"""
        return self.t_start + self.duration

    def _time_slice_indices(self, time_slice, strict_slicing=True):
        """
        Calculate the start and end indices for the slice.

        Also returns t_start
        """
        if time_slice is None:
            i_start, i_stop = None, None
            sig_t_start = self.t_start
        else:
            sr = self.sampling_rate
            t_start, t_stop = time_slice
            if t_start is None:
                i_start = None
                sig_t_start = self.t_start
            else:
                t_start = ensure_second(t_start)
                if strict_slicing:
                    if t_start < self.t_start or t_start > self.t_stop:
                        raise ValueError("`t_start` is outside of the time range")
                else:
                    t_start = max(t_start, self.t_start)
                # the i_start is rounded to the nearest sample
                i_start = np.rint((t_start - self.t_start).magnitude * sr.magnitude).astype(np.int64)
                # this is needed to get the real t_start of the first sample
                # because it does not necessary match what is demanded
                sig_t_start = self.t_start + i_start / sr

            if t_stop is None:
                i_stop = None
            else:
                t_stop = ensure_second(t_stop)
                if strict_slicing:
                    if t_stop < self.t_start or t_stop > self.t_stop:
                        raise ValueError("`t_stop` is outside of the time range")
                else:
                    t_stop = min(t_stop, self.t_stop)
                # calculate duration demanded then round it to nearest sample number
                # add this to i_start to get i_stop
                delta = (t_stop - t_start) * sr
                i_stop = i_start + int(np.rint(delta.simplified.magnitude))

        return i_start, i_stop, sig_t_start

    def load(self, time_slice=None, strict_slicing=True, channel_indexes=None, magnitude_mode="rescaled"):
        """
        *Args*:
            :time_slice: None or tuple of the time slice expressed with quantities.
                            None is the entire signal.
            :channel_indexes: None or list. Channels to load. None is all channels
                    Be carefull that channel_indexes represent the local channel index inside
                    the AnalogSignal and not the global_channel_indexes like in rawio.
            :magnitude_mode: 'rescaled' or 'raw'.
                For instance if the internal dtype is int16:
                    * **rescaled** give [1.,2.,3.]*pq.uV and the dtype is float32
                    * **raw** give [10, 20, 30]*pq.CompoundUnit('0.1*uV')
                The CompoundUnit with magnitude_mode='raw' is usefull to
                postpone the scaling when needed and having an internal dtype=int16
                but it less intuitive when you don't know so well quantities.
            :strict_slicing: True by default.
                Control if an error is raise or not when one of  time_slice member
                (t_start or t_stop) is outside the real time range of the segment.
        """

        # fixed_chan_indexes is channel index (or slice) in the stream
        # channel_indexes is channel index (or slice) in the substream
        if isinstance(self._inner_stream_channels, slice):
            if self._inner_stream_channels == slice(None):
                # sub stream is the entire stream
                if channel_indexes is None:
                    fixed_chan_indexes = None
                else:
                    fixed_chan_indexes = channel_indexes
            else:
                # sub stream is part of  stream with slice
                if channel_indexes is None:
                    fixed_chan_indexes = self._inner_stream_channels
                else:
                    global_inds = np.arange(self._nb_total_chann_in_stream)
                    fixed_chan_indexes = global_inds[self._inner_stream_channels][channel_indexes]
        else:
            # sub stream is part of  stream with indexes
            if channel_indexes is None:
                fixed_chan_indexes = self._inner_stream_channels
            else:
                fixed_chan_indexes = self._inner_stream_channels[channel_indexes]

        i_start, i_stop, sig_t_start = self._time_slice_indices(time_slice, strict_slicing=strict_slicing)

        raw_signal = self._rawio.get_analogsignal_chunk(
            block_index=self._block_index,
            seg_index=self._seg_index,
            i_start=i_start,
            i_stop=i_stop,
            stream_index=self._stream_index,
            channel_indexes=fixed_chan_indexes,
        )

        # if slice in channel : change name and array_annotations
        if raw_signal.shape[1] != self._nb_chan:
            name = "slice of  " + self.name
            channel_indexes2 = channel_indexes
            if channel_indexes2 is None:
                channel_indexes2 = slice(None)
            array_annotations = {k: v[channel_indexes2] for k, v in self.array_annotations.items()}
        else:
            name = self.name
            array_annotations = self.array_annotations

        if magnitude_mode == "raw":
            if self._raw_units is None:
                raise ValueError(
                    "raw magnitude is not supported if the gain are not the same for all channels or offset is not 0"
                )
            sig = raw_signal
            units = self._raw_units
        elif magnitude_mode == "rescaled":
            # dtype is float32 when internally it is float32 or int16
            if self.dtype == "float64":
                dtype = "float64"
            else:
                dtype = "float32"
            sig = self._rawio.rescale_signal_raw_to_float(
                raw_signal, dtype=dtype, stream_index=self._stream_index, channel_indexes=fixed_chan_indexes
            )
            units = self.units

        else:
            raise ValueError(f"Invalid magnitude_mode {magnitude_mode}. Accepted values are " f'"rescaled" and "raw"')

        anasig = AnalogSignal(
            sig,
            units=units,
            copy=False,
            t_start=sig_t_start,
            sampling_rate=self.sampling_rate,
            name=name,
            file_origin=self.file_origin,
            description=self.description,
            array_annotations=array_annotations,
            **self.annotations,
        )

        return anasig


class SpikeTrainProxy(BaseProxy):
    """
    This object mimic SpikeTrain except that it does not
    have the spike time nor waveforms.
    All attributes and annotations are here.

    The goal is to postpone the loading of data into memory
    when reading a file with the new lazy load system based
    on neo.rawio.

    This object must not be constructed directly but is given
    neo.io when lazy=True instead of a true SpikeTrain.

    The SpikeTrainProxy is able to load:
      * only a slice of time
      * load wveforms or not.
      * have an internal raw magnitude identic to the file (generally the ticks
        of clock in int64) or the rescale to seconds.

    Usage:
    >>> proxy_sptr = SpikeTrainProxy(rawio=self.reader, unit_channel=0,
                        block_index=0, seg_index=0,)
    >>> sptr = proxy_sptr.load()
    >>> slice_of_sptr = proxy_sptr.load(time_slice=(1.*pq.s, 2.*pq.s))

    """

    _parent_objects = ("Segment", "Unit")
    _quantity_attr = "times"
    _necessary_attrs = (("t_start", pq.Quantity, 0), ("t_stop", pq.Quantity, 0))
    _recommended_attrs = ()
    proxy_for = SpikeTrain

    def __init__(self, rawio=None, spike_channel_index=None, block_index=0, seg_index=0):

        self._rawio = rawio
        self._block_index = block_index
        self._seg_index = seg_index
        self._spike_channel_index = spike_channel_index

        nb_spike = self._rawio.spike_count(
            block_index=block_index, seg_index=seg_index, spike_channel_index=spike_channel_index
        )
        self.shape = (nb_spike,)

        self.t_start = self._rawio.segment_t_start(block_index, seg_index) * pq.s
        self.t_stop = self._rawio.segment_t_stop(block_index, seg_index) * pq.s

        seg_ann = self._rawio.raw_annotations["blocks"][block_index]["segments"][seg_index]
        annotations = seg_ann["spikes"][spike_channel_index].copy()
        array_annotations = annotations.pop("__array_annotations__")

        h = self._rawio.header["spike_channels"][spike_channel_index]
        wf_sampling_rate = h["wf_sampling_rate"]
        if not np.isnan(wf_sampling_rate) and wf_sampling_rate > 0:
            self.sampling_rate = wf_sampling_rate * pq.Hz
            self.left_sweep = (h["wf_left_sweep"] / self.sampling_rate).rescale("s")
            self._wf_units = ensure_signal_units(h["wf_units"])
        else:
            self.sampling_rate = None
            self.left_sweep = None

        BaseProxy.__init__(self, array_annotations=array_annotations, **annotations)

    def load(self, time_slice=None, strict_slicing=True, magnitude_mode="rescaled", load_waveforms=False):
        """
        *Args*:
            :time_slice: None or tuple of the time slice expressed with quantities.
                            None is the entire signal.
            :strict_slicing: True by default.
                 Control if an error is raise or not when one of  time_slice
                 member (t_start or t_stop) is outside the real time range of the segment.
            :magnitude_mode: 'rescaled' or 'raw'.
            :load_waveforms: bool load waveforms or not.
        """

        t_start, t_stop = consolidate_time_slice(time_slice, self.t_start, self.t_stop, strict_slicing)
        _t_start, _t_stop = prepare_time_slice(time_slice)

        spike_timestamps = self._rawio.get_spike_timestamps(
            block_index=self._block_index,
            seg_index=self._seg_index,
            spike_channel_index=self._spike_channel_index,
            t_start=_t_start,
            t_stop=_t_stop,
        )

        if magnitude_mode == "raw":
            # we must modify a bit the neo.rawio interface to also read the spike_timestamps
            # underlying clock wich is not always same as sigs
            raise (NotImplementedError)
        elif magnitude_mode == "rescaled":
            dtype = "float64"
            spike_times = self._rawio.rescale_spike_timestamp(spike_timestamps, dtype=dtype)
            units = "s"

        if load_waveforms:
            if self.sampling_rate is None:
                raise ValueError("There are no waveforms")
            raw_wfs = self._rawio.get_spike_raw_waveforms(
                block_index=self._block_index,
                seg_index=self._seg_index,
                spike_channel_index=self._spike_channel_index,
                t_start=_t_start,
                t_stop=_t_stop,
            )
            if magnitude_mode == "rescaled":
                float_wfs = self._rawio.rescale_waveforms_to_float(
                    raw_wfs, dtype="float32", spike_channel_index=self._spike_channel_index
                )
                waveforms = pq.Quantity(float_wfs, units=self._wf_units, dtype="float32", copy=False)
            elif magnitude_mode == "raw":
                # could code also CompundUnit here but it is over killed
                # so we used dimentionless
                waveforms = pq.Quantity(raw_wfs, units="", dtype=raw_wfs.dtype, copy=False)
        else:
            waveforms = None

        sptr = SpikeTrain(
            spike_times,
            t_stop,
            units=units,
            dtype=dtype,
            t_start=t_start,
            copy=False,
            sampling_rate=self.sampling_rate,
            waveforms=waveforms,
            left_sweep=self.left_sweep,
            name=self.name,
            file_origin=self.file_origin,
            description=self.description,
            **self.annotations,
        )

        if time_slice is None:
            sptr.array_annotate(**self.array_annotations)
        else:
            # TODO handle array_annotations with time_slice
            pass

        return sptr


class _EventOrEpoch(BaseProxy):
    _parent_objects = ("Segment",)
    _quantity_attr = "times"

    def __init__(self, rawio=None, event_channel_index=None, block_index=0, seg_index=0):

        self._rawio = rawio
        self._block_index = block_index
        self._seg_index = seg_index
        self._event_channel_index = event_channel_index

        nb_event = self._rawio.event_count(
            block_index=block_index, seg_index=seg_index, event_channel_index=event_channel_index
        )
        self.shape = (nb_event,)

        self.t_start = self._rawio.segment_t_start(block_index, seg_index) * pq.s
        self.t_stop = self._rawio.segment_t_stop(block_index, seg_index) * pq.s

        # both necessary attr and annotations
        annotations = {}
        for k in ("name", "id"):
            annotations[k] = self._rawio.header["event_channels"][event_channel_index][k]

        seg_ann = self._rawio.raw_annotations["blocks"][block_index]["segments"][seg_index]
        ann = seg_ann["events"][event_channel_index]
        annotations = ann.copy()
        array_annotations = annotations.pop("__array_annotations__")
        BaseProxy.__init__(self, array_annotations=array_annotations, **annotations)

    def load(self, time_slice=None, strict_slicing=True):
        """
        *Args*:
            :time_slice: None or tuple of the time slice expressed with quantities.
                            None is the entire signal.
            :strict_slicing: True by default.
                 Control if an error is raise or not when one of  time_slice member (t_start or t_stop)
                 is outside the real time range of the segment.
        """

        t_start, t_stop = consolidate_time_slice(time_slice, self.t_start, self.t_stop, strict_slicing)
        _t_start, _t_stop = prepare_time_slice(time_slice)

        timestamp, durations, labels = self._rawio.get_event_timestamps(
            block_index=self._block_index,
            seg_index=self._seg_index,
            event_channel_index=self._event_channel_index,
            t_start=_t_start,
            t_stop=_t_stop,
        )

        dtype = "float64"
        times = self._rawio.rescale_event_timestamp(timestamp, dtype=dtype)
        units = "s"

        if durations is not None:
            durations = self._rawio.rescale_epoch_duration(durations, dtype=dtype) * pq.s

        h = self._rawio.header["event_channels"][self._event_channel_index]
        if h["type"] == b"event":
            ret = Event(
                times=times,
                labels=labels,
                units="s",
                name=self.name,
                file_origin=self.file_origin,
                description=self.description,
                **self.annotations,
            )
        elif h["type"] == b"epoch":
            ret = Epoch(
                times=times,
                durations=durations,
                labels=labels,
                units="s",
                name=self.name,
                file_origin=self.file_origin,
                description=self.description,
                **self.annotations,
            )

        if time_slice is None:
            ret.array_annotate(**self.array_annotations)
        else:
            # TODO handle array_annotations with time_slice
            pass

        return ret


class EventProxy(_EventOrEpoch):
    """
    This object mimic Event except that it does not
    have the times nor labels.
    All other attributes and annotations are here.

    The goal is to postpone the loading of data into memory
    when reading a file with the new lazy load system based
    on neo.rawio.

    This object must not be constructed directly but is given
    neo.io when lazy=True instead of a true Event.

    The EventProxy is able to load:
      * only a slice of time

    Usage:
    >>> proxy_event = EventProxy(rawio=self.reader, event_channel_index=0,
                        block_index=0, seg_index=0,)
    >>> event = proxy_event.load()
    >>> slice_of_event = proxy_event.load(time_slice=(1.*pq.s, 2.*pq.s))

    """

    _necessary_attrs = (("times", pq.Quantity, 1), ("labels", np.ndarray, 1, np.dtype("U")))
    proxy_for = Event


class EpochProxy(_EventOrEpoch):
    """
    This object mimic Epoch except that it does not
    have the times nor labels nor durations.
    All other attributes and annotations are here.

    The goal is to postpone the loading of data into memory
    when reading a file with the new lazy load system based
    on neo.rawio.

    This object must not be constructed directly but is given
    neo.io when lazy=True instead of a true Epoch.

    The EpochProxy is able to load:
      * only a slice of time

    Usage:
    >>> proxy_epoch = EpochProxy(rawio=self.reader, event_channel_index=0,
                        block_index=0, seg_index=0,)
    >>> epoch = proxy_epoch.load()
    >>> slice_of_epoch = proxy_epoch.load(time_slice=(1.*pq.s, 2.*pq.s))

    """

    _necessary_attrs = (
        ("times", pq.Quantity, 1),
        ("durations", pq.Quantity, 1),
        ("labels", np.ndarray, 1, np.dtype("U")),
    )
    proxy_for = Epoch


proxyobjectlist = [AnalogSignalProxy, SpikeTrainProxy, EventProxy, EpochProxy]


unit_convert = {
    "Volts": "V",
    "volts": "V",
    "Volt": "V",
    "volt": "V",
    " Volt": "V",
    "microV": "uV",
    # note that "micro" and "mu" are two different characters in Unicode
    # although they mostly look the same. Here we accept both.
    "µV": "uV",
    "μV": "uV",
}


def ensure_signal_units(units):
    # test units
    units = units.replace(" ", "")
    if units in unit_convert:
        units = unit_convert[units]
    try:
        units = pq.Quantity(1, units)
    except:
        logger.warning(f'Units "{units}" can not be converted to a quantity. Using dimensionless ' "instead")
        units = ""
        units = pq.Quantity(1, units)
    return units


def check_annotations(annotations):
    # force type to str for some keys
    # imposed for tests
    for k in ("name", "description", "file_origin"):
        if k in annotations:
            annotations[k] = str(annotations[k])

    if "coordinates" in annotations:
        # some rawio expose some coordinates in annotations but is not standardized
        # (x, y, z) or polar, at the moment it is more resonable to remove them
        annotations.pop("coordinates")

    return annotations


def ensure_second(v):
    if isinstance(v, float):
        return v * pq.s
    elif isinstance(v, pq.Quantity):
        return v.rescale("s")
    elif isinstance(v, int):
        return float(v) * pq.s


def prepare_time_slice(time_slice):
    """
    This give clean time slice but keep None
    for calling rawio slice
    """
    if time_slice is None:
        t_start, t_stop = None, None
    else:
        t_start, t_stop = time_slice

    if t_start is not None:
        t_start = ensure_second(t_start).rescale("s").magnitude

    if t_stop is not None:
        t_stop = ensure_second(t_stop).rescale("s").magnitude

    return (t_start, t_stop)


def consolidate_time_slice(time_slice, seg_t_start, seg_t_stop, strict_slicing):
    """
    This give clean time slice in quantity for t_start/t_stop of object
    None is replace by seg limit.
    """
    if time_slice is None:
        t_start, t_stop = None, None
    else:
        t_start, t_stop = time_slice

    if t_start is None:
        t_start = seg_t_start
    else:
        if strict_slicing:
            if t_start < seg_t_start or t_start > seg_t_stop:
                raise ValueError("`t_start` is outside of the time range")
        else:
            t_start = max(t_start, seg_t_start)
    t_start = ensure_second(t_start)

    if t_stop is None:
        t_stop = seg_t_stop
    else:
        if strict_slicing:
            if t_stop < seg_t_start or t_stop > seg_t_stop:
                raise ValueError("`t_stop` is outside of the time range")
        else:
            t_stop = min(t_stop, seg_t_stop)
    t_stop = ensure_second(t_stop)

    return (t_start, t_stop)



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.spike2rawio import Spike2RawIO


class Spike2IO(Spike2RawIO, BaseFromRaw):
    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename, **kargs):
        Spike2RawIO.__init__(self, filename=filename, **kargs)
        BaseFromRaw.__init__(self, filename)



"""
Reading files with neo.io
=========================

"""

####################################################
# Start with package import and getting a datafile

import urllib

import neo

url_repo = "https://web.gin.g-node.org/NeuralEnsemble/ephy_testing_data/raw/master/"

# Plexon files
distantfile = url_repo + "plexon/File_plexon_3.plx"
localfile = "File_plexon_3.plx"
urllib.request.urlretrieve(distantfile, localfile)


###################################################
# Now we can create our reader and read some data

# create a reader
reader = neo.io.PlexonIO(filename="File_plexon_3.plx")
# read the blocks
blks = reader.read(lazy=False)
print(blks)

######################################################
# Once we have our blocks we can iterate through each
# block of data and see the contents of all parts of
# that data

# access to segments
for blk in blks:
    for seg in blk.segments:
        print(seg)
        for asig in seg.analogsignals:
            print(asig)
        for st in seg.spiketrains:
            print(st)

#######################################################
# Let's look at another file type

# CED Spike2 files
distantfile = url_repo + "spike2/File_spike2_1.smr"
localfile = "./File_spike2_1.smr"
urllib.request.urlretrieve(distantfile, localfile)
# create a reader
reader = neo.io.Spike2IO(filename="File_spike2_1.smr")

#########################################################
# Despite being a different raw file format we can access
# the data in the same way

# read the block
bl = reader.read(lazy=False)[0]
print(bl)

##########################################################
# Similarly we can view the different types of data within
# the block (AnalogSignals and SpikeTrains)

# access to segments
for seg in bl.segments:
    print(seg)
    for asig in seg.analogsignals:
        print(asig)
    for st in seg.spiketrains:
        print(st)
